from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from typing import Final, Literal, TypedDict, cast

import e3nn
import e3nn.nn
import e3nn.util.jit
import torch
from e3nn import o3
from torch import Tensor

from graph_pes.atomic_graph import (
    DEFAULT_CUTOFF,
    AtomicGraph,
    PropertyKey,
    index_over_neighbours,
    neighbour_distances,
    neighbour_vectors,
)
from graph_pes.graph_pes_model import GraphPESModel
from graph_pes.models.components import distances
from graph_pes.models.components.aggregation import (
    NeighbourAggregation,
    NeighbourAggregationMode,
)
from graph_pes.models.components.scaling import LocalEnergiesScaler
from graph_pes.models.e3nn.utils import (
    LinearReadOut,
    SphericalHarmonics,
    build_limited_tensor_product,
)
from graph_pes.utils.logger import logger
from graph_pes.utils.nn import (
    MLP,
    AtomicOneHot,
    HaddamardProduct,
    PerElementEmbedding,
    UniformModuleList,
)

warnings.filterwarnings(
    "ignore",
    message=(
        "The TorchScript type system doesn't support instance-level "
        "annotations on empty non-base types.*"
    ),
)

### IMPLEMENTATION ###


def build_gate(output_irreps: o3.Irreps):
    """
    Builds an equivariant non-linearity that produces ``output_irreps``.

    This is done by passing all scalar irreps directly through a non-linearity,
    (tanh for even parity, silu for odd parity).

    All higher order irreps are multiplied by a suitable number of scalar
    irrep values that have also been passed through a non-linearity.

    The gated scalar and higher order irreps are then combined to give the
    final output.
    """

    parity_to_activations = {
        -1: torch.nn.functional.tanh,
        1: torch.nn.functional.silu,
    }

    # separate scalar and higher order irreps
    scalar_inputs = o3.Irreps(
        irreps for irreps in output_irreps if irreps.ir.l == 0
    )
    scalar_activations = [
        parity_to_activations[irrep.ir.p] for irrep in scalar_inputs
    ]

    higher_inputs = o3.Irreps(
        irreps for irreps in output_irreps if irreps.ir.l > 0
    )

    # work out the the `irrep_gates` scalars to be used
    gate_irrep = "0e" if "0e" in output_irreps else "0o"
    gate_parity = o3.Irrep(gate_irrep).p
    gate_value_irreps = o3.Irreps(
        (channels, gate_irrep) for (channels, _) in higher_inputs
    )
    gate_activations = [
        parity_to_activations[gate_parity] for _ in higher_inputs
    ]

    return e3nn.nn.Gate(
        irreps_scalars=scalar_inputs,
        act_scalars=scalar_activations,
        irreps_gates=gate_value_irreps,
        act_gates=gate_activations,
        irreps_gated=higher_inputs,
    )


class SelfInteraction(ABC, torch.nn.Module):
    @abstractmethod
    def forward(self, node_embeddings: Tensor, Z_embeddings: Tensor) -> Tensor:
        pass


class LinearSelfInteraction(SelfInteraction):
    def __init__(
        self,
        input_irreps: o3.Irreps,
        output_irreps: o3.Irreps,
    ):
        super().__init__()

        self.linear = o3.Linear(input_irreps, output_irreps)

    def forward(self, node_embeddings: Tensor, Z_embeddings: Tensor) -> Tensor:
        return self.linear(node_embeddings)


class TensorProductSelfInteraction(SelfInteraction):
    def __init__(
        self,
        input_irreps: o3.Irreps | str,
        output_irreps: o3.Irreps | str,
        Z_embedding_irreps: o3.Irreps | str,
    ):
        super().__init__()

        self.tensor_product = o3.FullyConnectedTensorProduct(
            input_irreps,
            Z_embedding_irreps,
            output_irreps,
        )

    def forward(self, node_embeddings: Tensor, Z_embeddings: Tensor) -> Tensor:
        return self.tensor_product(node_embeddings, Z_embeddings)


class NequIPMessagePassingLayer(torch.nn.Module):
    def __init__(
        self,
        input_node_irreps: o3.Irreps,
        Z_embedding_dim: int,
        radial_features: int,
        edge_features: o3.Irreps,
        target_node_features: o3.Irreps,
        cutoff: float,
        self_interaction: Literal["linear", "tensor_product"] | None,
        prune_output_to: list[o3.Irrep] | None,
        neighbour_aggregation: NeighbourAggregationMode,
    ):
        super().__init__()

        target_multiplicities: dict[o3.Irrep, int] = {}
        for mul, ir in target_node_features:
            target_multiplicities[ir] = mul

        # NequIP message passing involves the following steps:
        # 1. Create the message from neighbour j to central atom i
        #   - update neighbour node features, h_j, with a linear layer
        #   - mix the neighbour node features with the edge features, using a
        #       tensor product with weights that are learned as a function
        #       of the neighbour distance, to create messages m_j_to_i
        # 2. Aggregate the messages over all neighbours to create the
        #       total message, m_i
        # 3. Pass the total message through a linear layer to create
        #       new central atom embeddings
        # 4. (if configured) Allow the old central atom embeddings to interact
        #       with themselves via another tensor product and residual add
        # 5. Apply a non-linearity to the new central atom embeddings

        # 1. Message creation
        self.pre_message_linear = o3.Linear(
            input_node_irreps, input_node_irreps
        )
        self.message_tensor_product = build_limited_tensor_product(
            node_embedding_irreps=input_node_irreps,
            edge_embedding_irreps=edge_features,
            allowed_outputs=[ir for _, ir in target_node_features],
        )
        n_required_weights = self.message_tensor_product.weight_numel
        self.weight_generator = HaddamardProduct(
            torch.nn.Sequential(
                distances.Bessel(radial_features, cutoff, trainable=True),
                MLP(
                    [radial_features] * 3 + [n_required_weights],
                    activation=torch.nn.SiLU(),
                ),
            ),
            distances.PolynomialEnvelope(cutoff, p=6),
        )

        # 2. Message aggregation
        self.aggregation = NeighbourAggregation.parse(
            mode=neighbour_aggregation
        )

        # 5. Non-linearity
        # NB we initialise this first so we can work out how many irreps
        #    we need to be generating in step 3
        post_message_irreps: list[o3.Irrep] = sorted(
            set(i.ir for i in self.message_tensor_product.irreps_out)
        )
        if not prune_output_to:
            # desired_output_irreps = post_message_irreps
            irreps = []
            for ir in post_message_irreps:
                irreps.append((target_multiplicities[ir], ir))
            desired_output_irreps = o3.Irreps(irreps)
        else:
            irreps = []
            for ir in post_message_irreps:
                if ir in prune_output_to:
                    irreps.append((target_multiplicities[ir], ir))
            desired_output_irreps = o3.Irreps(irreps)

        self.non_linearity = build_gate(desired_output_irreps)  # type: ignore

        # 3. Post-message linear
        self.post_message_linear = o3.Linear(
            self.message_tensor_product.irreps_out.simplify(),
            self.non_linearity.irreps_in,
        )

        # 4. Self-interaction
        # create a self interaction that produces the output irreps that
        # are required for the non-linearity
        self.self_interaction: SelfInteraction | None = None

        if self_interaction == "tensor_product":
            self.self_interaction = TensorProductSelfInteraction(
                input_irreps=input_node_irreps,
                Z_embedding_irreps=f"{Z_embedding_dim}x0e",
                output_irreps=self.non_linearity.irreps_in,
            )
        elif self_interaction == "linear":
            self.self_interaction = LinearSelfInteraction(
                input_irreps=input_node_irreps,
                output_irreps=self.non_linearity.irreps_in,
            )

        # bookkeeping
        self.irreps_in = input_node_irreps
        self.irreps_out: o3.Irreps = self.non_linearity.irreps_out

    def forward(
        self,
        node_embeddings: Tensor,  # [n_atoms, irreps_in]
        Z_embeddings: Tensor,  # [n_atoms, Z_embedding_dim]
        neighbour_distances: Tensor,  # [n_edges,]
        edge_embedding: Tensor,  # [n_edges, edge_irreps]
        graph: AtomicGraph,
    ) -> Tensor:  # [n_atoms, irreps_out]
        # 1. message creation
        neighbour_embeddings = index_over_neighbours(
            self.pre_message_linear(node_embeddings), graph
        )
        weights = self.weight_generator(neighbour_distances.unsqueeze(-1))
        messages = self.message_tensor_product(
            neighbour_embeddings, edge_embedding, weights
        )

        # 2. message aggregation
        total_message = self.aggregation(messages, graph)

        # 3. update node embeddings
        new_node_embeddings = self.post_message_linear(total_message)

        # 4. self-interaction
        if self.self_interaction is not None:
            new_node_embeddings = new_node_embeddings + self.self_interaction(
                node_embeddings, Z_embeddings
            )

        # 5. non-linearity
        return self.non_linearity(new_node_embeddings)

    # type hints for mypy
    def __call__(
        self,
        node_embeddings: Tensor,
        Z_embeddings: Tensor,
        neighbour_distances: Tensor,
        edge_embedding: Tensor,
        graph: AtomicGraph,
    ) -> Tensor:
        return super().__call__(
            node_embeddings,
            Z_embeddings,
            neighbour_distances,
            edge_embedding,
            graph,
        )


@e3nn.util.jit.compile_mode("script")
class _BaseNequIP(GraphPESModel):
    def __init__(
        self,
        # model
        cutoff: float,
        direct_force_predictions: bool,
        # input embeddings
        Z_embedding: torch.nn.Module,
        Z_embedding_dim: int,
        radial_features: int,
        # message passing
        layers: int,
        node_features: o3.Irreps,
        edge_features: o3.Irreps,
        self_interaction: Literal["linear", "tensor_product"] | None,
        neighbour_aggregation: NeighbourAggregationMode,
        # optimisation
        prune_last_layer: bool,
    ):
        props: list[PropertyKey] = ["local_energies"]
        if direct_force_predictions:
            props.append("forces")

        super().__init__(
            cutoff=cutoff,
            implemented_properties=props,
        )

        if not prune_last_layer:
            prune_output_to = None
        else:
            if direct_force_predictions:
                prune_output_to = [o3.Irrep("0e"), o3.Irrep("1o")]
            else:
                prune_output_to = [o3.Irrep("0e")]

        self.Z_embedding = Z_embedding

        scalar_even_dim = node_features.count("0e")
        if scalar_even_dim == 0:
            raise ValueError("Hidden irreps must contain a `0e` component.")
        self.initial_node_embedding = PerElementEmbedding(scalar_even_dim)
        # l_max = max(ir.l for ir in hidden_irreps)
        # use_odd_parity = any(ir.ir.p == -1 for ir in hidden_irreps)

        self.edge_embedding = SphericalHarmonics(edge_features, normalize=True)

        # first layer recieves an even parity, scalar embedding
        # of the atomic number
        current_layer_input: o3.Irreps
        current_layer_input = o3.Irreps(f"{scalar_even_dim}x0e")  # type: ignore
        _layers: list[NequIPMessagePassingLayer] = []
        for i in range(layers):
            layer = NequIPMessagePassingLayer(
                input_node_irreps=current_layer_input,
                edge_features=edge_features,
                target_node_features=node_features,
                cutoff=cutoff,
                radial_features=radial_features,
                Z_embedding_dim=Z_embedding_dim,
                self_interaction=self_interaction,
                prune_output_to=None if i < layers - 1 else prune_output_to,
                neighbour_aggregation=neighbour_aggregation,
            )
            _layers.append(layer)
            current_layer_input = layer.irreps_out

        self.layers = UniformModuleList(_layers)
        self.energy_readout = LinearReadOut(current_layer_input)

        if direct_force_predictions:
            if current_layer_input.count("1o") == 0:
                raise ValueError(
                    "Hidden irreps must contain a `1o` component in order "
                    f"to predict forces. Got {current_layer_input}."
                )
            self.force_readout = LinearReadOut(current_layer_input, "1o")

        else:
            self.force_readout = None

        self.scaler = LocalEnergiesScaler()

    def forward(self, graph: AtomicGraph) -> dict[PropertyKey, torch.Tensor]:
        # pre-compute important quantities
        r = neighbour_distances(graph)
        Y = self.edge_embedding(neighbour_vectors(graph))
        Z_embed = self.Z_embedding(graph.Z)

        # initialise the node embeddings...
        node_embed = self.initial_node_embedding(graph.Z)

        # ...iteratively update them...
        for layer in self.layers:
            node_embed = layer(node_embed, Z_embed, r, Y, graph)

        # ...and read out
        local_energies = self.energy_readout(node_embed).squeeze()
        local_energies = self.scaler(local_energies, graph)

        preds: dict[PropertyKey, torch.Tensor] = {
            "local_energies": local_energies
        }
        if self.force_readout is not None:
            preds["forces"] = self.force_readout(node_embed).squeeze()

        return preds


### USER FACING INTERFACE ###


class SimpleIrrepSpec(TypedDict):
    r"""
    A simple specification of the node and edge feature irreps for
    :class:`~graph_pes.models.NequIP`.

    Parameters
    ----------
    channels
        The number of channels for the node embedding. If an integer, all
        :math:`l`-order irreps will have the same number of channels. If a list,
        the :math:`l`-order irreps will have the number of channels specified by
        the :math:`l`-th entry in the list.
    l_max
        The maximum angular momentum for the edge embedding.
    use_odd_parity
        Whether to allow odd parity for the edge embedding.

    Examples
    --------

    .. code:: python

        >>> from graph_pes.models.e3nn.nequip import SimpleIrrepSpec
        >>> SimpleIrrepSpec(channels=16, l_max=2, use_odd_parity=True)
    """

    channels: int | list[int]
    l_max: int
    use_odd_parity: bool


class CompleteIrrepSpec(TypedDict):
    r"""
    A complete specification of the node and edge feature irreps for
    :class:`~graph_pes.models.NequIP`.

    Parameters
    ----------
    node_irreps
        The node feature irreps.
    edge_irreps
        The edge feature irreps.

    Examples
    --------

    .. code:: python

        >>> from graph_pes.models.e3nn.nequip import CompleteIrrepSpec
        >>> CompleteIrrepSpec(
        ...     node_irreps="32x0e + 16x1o + 8x2e",
        ...     edge_irreps="0e + 1o + 2e"
        ... )
    """

    node_irreps: str
    edge_irreps: str


DEFAULT_FEATURES: Final[SimpleIrrepSpec] = {
    "channels": 16,
    "l_max": 2,
    "use_odd_parity": True,
}


def parse_irrep_specification(
    spec: SimpleIrrepSpec | CompleteIrrepSpec,
) -> tuple[o3.Irreps, o3.Irreps]:
    is_simple = set(spec.keys()) == {"channels", "l_max", "use_odd_parity"}
    is_complete = set(spec.keys()) == {"node_irreps", "edge_irreps"}

    if not is_simple and not is_complete:
        raise ValueError(
            "Invalid irrep specification. Expected a dict with keys "
            "`node_irreps` and `edge_irreps` or a dict with keys "
            "`channels`, `l_max`, and `use_odd_parity`."
        )

    if is_simple:
        spec = cast(SimpleIrrepSpec, spec)
        l_max, channels = spec["l_max"], spec["channels"]
        if not isinstance(channels, list):
            channels = [channels] * (l_max + 1)

        if len(channels) != l_max + 1:
            raise ValueError(
                "channels must be an integer or a list of length l_max + 1"
            )

        parities = "oe" if spec["use_odd_parity"] else "e"

        node_irreps = []
        for l, c in zip(range(l_max + 1), channels):
            for p in parities:
                node_irreps.append(f"{c}x{l}{p}")

        edge_irreps = []
        for l in range(l_max + 1):
            p = "e" if not spec["use_odd_parity"] else "eo"[l % 2]
            edge_irreps.append(f"1x{l}{p}")

        node_irreps = o3.Irreps(" + ".join(node_irreps))
        edge_irreps = o3.Irreps(" + ".join(edge_irreps))

    else:
        spec = cast(CompleteIrrepSpec, spec)
        node_irreps = o3.Irreps(spec["node_irreps"])
        edge_irreps = o3.Irreps(spec["edge_irreps"])

    logger.debug(
        f"""\
        Parsed NequIP irrep speficication:
            node_irreps: {node_irreps}
            edge_irreps: {edge_irreps}
        """
    )

    return node_irreps, edge_irreps  # type: ignore


@e3nn.util.jit.compile_mode("script")
class NequIP(_BaseNequIP):
    r"""
    NequIP architecture from `E(3)-equivariant graph neural networks for
    data-efficient and accurate interatomic potentials
    <https://www.nature.com/articles/s41467-022-29939-5>`__.

    If you use this model in your research, please cite the original work:

    .. code:: bibtex

        @article{Batzner-22-05,
            title = {
                      E(3)-Equivariant Graph Neural Networks for
                      Data-Efficient and Accurate Interatomic Potentials
                    },
            author = {
                      Batzner, Simon and Musaelian, Albert and Sun, Lixin
                      and Geiger, Mario and Mailoa, Jonathan P.
                      and Kornbluth, Mordechai and Molinari, Nicola
                      and Smidt, Tess E. and Kozinsky, Boris
                    },
            year = {2022},
            journal = {Nature Communications},
            volume = {13},
            number = {1},
            pages = {2453},
            doi = {10.1038/s41467-022-29939-5},
            copyright = {2022 The Author(s)}
        }

    Parameters
    ----------
    elements
        The elements that the model will encounter in the training data.
        This is used to create the atomic one-hot embedding. If you intend
        to fine-tune this model on additional data, you must ensure that all
        elements you will encounter in both pre-training and fine-tuning are
        present in this list. (See :class:`~graph_pes.models.ZEmbeddingNequIP`
        for an alternative that allows for arbitrary atomic numbers.)
    direct_force_predictions
        Whether to predict forces directly. If ``True``, the model will output
        forces (rather than infer them from the energy) using a
        :class:`~graph_pes.models.e3nn.utils.LinearReadOut` to map the final
        layer node embedding to a set of force predictions.
    cutoff
        The cutoff radius for the model.
    features
        A specification of the irreps to use for the node and edge embeddings.
        Can be either a SimpleIrrepSpec or a CompleteIrrepSpec.
    layers
        The number of layers for the message passing.
    self_interaction
        The kind of self-interaction to use. If ``None``, no self-interaction
        is applied. If ``"linear"``, a linear self-interaction is applied
        to update the node embeddings along the residual path. If
        ``"tensor_product"``, a tensor product combining the old node
        embedding with an embedding of the atomic number is applied.
        As first noticed by the authors of `SevenNet <https://pubs.acs.org/doi/10.1021/acs.jctc.4c00190>`__,
        using linear self-interactions greatly reduces the number of parameters
        in the model, and helps to prevent overfitting.
    prune_last_layer
        Whether to prune irrep communication pathways in the final layer
        that do not contribute to the ``0e`` output embedding.
    neighbour_aggregation
        The neighbour aggregation mode. See
        :class:`~graph_pes.models.components.aggregation.NeighbourAggregationMode`
        for more details. Note that ``"mean"`` or ``"sqrt"`` aggregations lead
        to un-physical discontinuities in the energy function as atoms enter and
        leave the cutoff radius of the model.
    radial_features
        The number of features to expand the radial distances into. These
        features are then passed through an :class:`~graph_pes.utils.nn.MLP` to
        generate distance-conditioned weights for the message tensor product.

    Examples
    --------

    Configure a NequIP model for use with
    :doc:`graph-pes-train <../../cli/graph-pes-train/root>`:

    .. code:: yaml

        model:
          +NequIP:
            elements: [C, H, O]
            cutoff: 5.0

            # use 2 message passing layers
            layers: 2

            # using SimpleIrrepSpec
            features:
              channels: [64, 32, 8]
              l_max: 2
              use_odd_parity: true

            # scale the aggregation by the avg. number of
            # neighbours in the training set
            neighbour_aggregation: constant_fixed

    The hidden layer and edge embedding irreps the model generates can be
    controlled using either a :class:`~graph_pes.models.e3nn.nequip.SimpleIrrepSpec`
    or a :class:`~graph_pes.models.e3nn.nequip.CompleteIrrepSpec`:

    .. code:: python

        >>> from graph_pes.models import NequIP
        >>> model = NequIP(
        ...     elements=["C", "H", "O"],
        ...     cutoff=5.0,
        ...     features={
        ...         "channels": [16, 8, 4],
        ...         "l_max": 2,
        ...         "use_odd_parity": True
        ...     },
        ...     layers=3,
        ... )
        >>> for layer in model.layers:
        ...     print(layer.irreps_in, "->", layer.irreps_out)
        16x0e -> 16x0e+8x1o+4x2e
        16x0e+8x1o+4x2e -> 16x0e+8x1o+8x1e+4x2o+4x2e
        16x0e+8x1o+8x1e+4x2o+4x2e -> 16x0e

    .. code:: python

        >>> from graph_pes.models import NequIP
        >>> model = NequIP(
        ...     elements=["C", "H", "O"],
        ...     cutoff=5.0,
        ...     features={
        ...         "node_irreps": "32x0e + 16x1o + 8x2e",
        ...         "edge_irreps": "1x0e + 1x1o + 1x2e"
        ...     },
        ...     layers=3,
        ... )
        >>> for layer in model.layers:
        ...     print(layer.irreps_in, "->", layer.irreps_out)
        32x0e -> 32x0e+16x1o+8x2e
        32x0e+16x1o+8x2e -> 32x0e+16x1o+8x2e
        32x0e+16x1o+8x2e -> 32x0e

    Observe the drop in parameters as we prune the last layer and
    replace the tensor product interactions with linear layers:

    .. code:: python

        >>> from graph_pes.models import NequIP
        >>> # vanilla NequIP
        >>> vanilla = NequIP(
        ...     elements=["C", "H", "O"],
        ...     cutoff=5.0,
        ...     features={"channels": 128, "l_max": 2, "use_odd_parity": True},
        ...     layers=3,
        ...     self_interaction="tensor_product",
        ...     prune_last_layer=False,
        ... )
        >>> sum(p.numel() for p in vanilla.parameters())
        2308720
        >>> # SevenNet-flavoured NequIP
        >>> smaller = NequIP(
        ...     elements=["C", "H", "O"],
        ...     cutoff=5.0,
        ...     features={"channels": 128, "l_max": 2, "use_odd_parity": True},
        ...     layers=3,
        ...     self_interaction="linear",
        ...     prune_last_layer=True,
        ... )
        >>> sum(p.numel() for p in smaller.parameters())
        965232
    """  # noqa: E501

    def __init__(
        self,
        elements: list[str],
        direct_force_predictions: bool = False,
        cutoff: float = DEFAULT_CUTOFF,
        layers: int = 3,
        features: SimpleIrrepSpec | CompleteIrrepSpec = DEFAULT_FEATURES,
        self_interaction: Literal["linear", "tensor_product"]
        | None = "tensor_product",
        prune_last_layer: bool = True,
        neighbour_aggregation: NeighbourAggregationMode = "sum",
        radial_features: int = 8,
    ):
        assert len(set(elements)) == len(elements), "Found duplicate elements"
        Z_embedding = AtomicOneHot(elements)
        Z_embedding_dim = len(elements)

        node_features, edge_features = parse_irrep_specification(features)

        super().__init__(
            direct_force_predictions=direct_force_predictions,
            Z_embedding=Z_embedding,
            Z_embedding_dim=Z_embedding_dim,
            node_features=node_features,
            edge_features=edge_features,
            layers=layers,
            cutoff=cutoff,
            self_interaction=self_interaction,
            prune_last_layer=prune_last_layer,
            neighbour_aggregation=neighbour_aggregation,
            radial_features=radial_features,
        )


@e3nn.util.jit.compile_mode("script")
class ZEmbeddingNequIP(_BaseNequIP):
    r"""
    A modified version of the :class:`~graph_pes.models.NequIP` architecture
    that embeds atomic numbers into a learnable embedding rather than using an
    atomic one-hot encoding.

    This circumvents the need to know in advance all the elements that the
    model will encounter in any pre-training or fine-tuning datasets.

    **Relevant differences** from :class:`~graph_pes.models.NequIP`\ :

    - The ``elements`` argument is removed.
    - The ``Z_embed_dim`` (:class:`int`) argument controls the size of the
      atomic number embedding (default: 8).

    For all other options, see :class:`~graph_pes.models.NequIP`.
    """

    def __init__(
        self,
        cutoff: float = DEFAULT_CUTOFF,
        direct_force_predictions: bool = False,
        Z_embed_dim: int = 8,
        features: SimpleIrrepSpec | CompleteIrrepSpec = DEFAULT_FEATURES,
        layers: int = 3,
        self_interaction: Literal["linear", "tensor_product"]
        | None = "tensor_product",
        prune_last_layer: bool = True,
        neighbour_aggregation: NeighbourAggregationMode = "sum",
        radial_features: int = 8,
    ):
        Z_embedding = PerElementEmbedding(Z_embed_dim)
        node_features, edge_features = parse_irrep_specification(features)
        super().__init__(
            direct_force_predictions=direct_force_predictions,
            Z_embedding=Z_embedding,
            Z_embedding_dim=Z_embed_dim,
            node_features=node_features,
            edge_features=edge_features,
            layers=layers,
            cutoff=cutoff,
            self_interaction=self_interaction,
            prune_last_layer=prune_last_layer,
            neighbour_aggregation=neighbour_aggregation,
            radial_features=radial_features,
        )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Callable, Literal, NamedTuple, Sequence

import torch
from torch import Tensor, nn

from graph_pes.atomic_graph import AtomicGraph, PropertyKey, divide_per_atom
from graph_pes.graph_pes_model import GraphPESModel
from graph_pes.utils.misc import uniform_repr
from graph_pes.utils.nn import UniformModuleList

Metric = Callable[[Tensor, Tensor], Tensor]
MetricName = Literal["MAE", "RMSE", "MSE"]


class WeightedLoss:
    def __init__(self, *args, **kwargs):
        # this is now depracated: pass weight directly to the loss object
        raise ImportError(
            "The WeightedLoss class has been removed from graph-pes "
            "as of version 0.0.22. Please now pass loss weights directly "
            "to the loss instances! See the docs for more information: "
            "https://jla-gardner.github.io/graph-pes/fitting/losses.html"
        )


class Loss(nn.Module, ABC):
    """
    A general base class for all loss functions in ``graph-pes``.

    Implementations **must** override:

    * :meth:`forward` to compute the loss value.
    * :meth:`name` to return the name of the loss function.
    * :meth:`required_properties` to return the properties that this loss
      function needs to have available in order to compute its value.

    Additionally, implementations can optionally override:

    * :meth:`pre_fit` to perform any necessary operations before training
      commences.

    Parameters
    ----------
    weight
        A scalar multiplier for weighting the value returned by
        :meth:`forward` as part of a :class:`TotalLoss`.
    """

    def __init__(self, weight):
        super().__init__()
        self.weight = weight

    @abstractmethod
    def forward(
        self,
        model: GraphPESModel,
        graph: AtomicGraph,
        predictions: dict[PropertyKey, torch.Tensor],
    ) -> torch.Tensor:
        r"""
        Compute the unweighted loss value.

        :class:`Loss`\ s can act on any of:

        Parameters
        ----------
        model
            The model being trained.
        graph
            The graph (usually a batch) the ``model`` was applied to.
        predictions
            The predictions from the ``model`` for the given ``graph``.
        """

    @property
    @abstractmethod
    def required_properties(self) -> list[PropertyKey]:
        """The properties that are required by this loss function."""

    @property
    @abstractmethod
    def name(self) -> str:
        """The name of this loss function, for logging purposes."""

    def pre_fit(self, training_data: AtomicGraph):
        """
        Perform any necessary operations before training commences.

        For example, this could be used to pre-compute a standard deviation
        of some property in the training data, which could then be used in
        :meth:`forward`.

        Parameters
        ----------
        training_data
            The training data to pre-fit this loss function to.
        """

    # add type hints to play nicely with mypy
    def __call__(
        self,
        model: GraphPESModel,
        graph: AtomicGraph,
        predictions: dict[PropertyKey, torch.Tensor],
    ) -> torch.Tensor:
        return super().__call__(model, graph, predictions)


class PropertyLoss(Loss):
    r"""
    A :class:`PropertyLoss` instance applies its :class:`Metric` to compare a
    model's predictions to the true values for a given property of a
    :class:`~graph_pes.AtomicGraph`.

    Parameters
    ----------
    property
        The property to apply the loss metric to.
    metric
        The loss metric to use. Defaults to :class:`RMSE`.

    Examples
    --------

    .. code-block:: python

        energy_rmse_loss = PropertyLoss("energy", RMSE())
        energy_rmse_value = energy_rmse_loss(
            predictions,  # a dict of key (energy/force/etc.) to value
            graph.properties,
        )
    """

    def __init__(
        self,
        property: PropertyKey,
        metric: Metric | MetricName = "RMSE",
        weight: float = 1.0,
    ):
        super().__init__(weight)
        self.property: PropertyKey = property
        self.metric = parse_metric(metric)

    def forward(
        self,
        model: GraphPESModel,
        graph: AtomicGraph,
        predictions: dict[PropertyKey, torch.Tensor],
    ) -> torch.Tensor:
        """
        Computes the loss value.

        Parameters
        ----------
        predictions
            The predictions from the ``model`` for the given ``graph``.
        """

        return self.metric(
            predictions[self.property],
            graph.properties[self.property],
        )

    @property
    def name(self) -> str:
        """Get the name of this loss for logging purposes."""
        return f"{self.property}_{_get_metric_name(self.metric)}"

    @property
    def required_properties(self) -> list[PropertyKey]:
        return [self.property]

    def __repr__(self) -> str:
        return uniform_repr(
            self.__class__.__name__,
            self.property,
            metric=self.metric,
        )


class SubLossPair(NamedTuple):
    loss_value: torch.Tensor
    weighted_loss_value: torch.Tensor


class TotalLossResult(NamedTuple):
    loss_value: torch.Tensor
    components: dict[str, SubLossPair]


class TotalLoss(torch.nn.Module):
    r"""
    A lightweight wrapper around a collection of losses.

    .. math::

        \mathcal{L}_{\text{total}} = \sum_i w_i \mathcal{L}_i

    where :math:`\mathcal{L}_i` is the :math:`i`-th loss and :math:`w_i` is the
    corresponding weight.

    ``graph-pes`` models are trained by minimising a :class:`TotalLoss` value.

    Parameters
    ----------
    losses
        The collection of losses to aggregate.
    """

    def __init__(self, losses: Sequence[Loss]):
        super().__init__()
        self.losses = UniformModuleList(losses)

    def forward(
        self,
        model: GraphPESModel,
        graph: AtomicGraph,
        predictions: dict[PropertyKey, torch.Tensor],
    ) -> TotalLossResult:
        """
        Computes the total loss value.

        Parameters
        ----------
        predictions
            The predictions from the model.
        graph
            The graph (usually a batch) the ``model`` was applied to.
        """

        total_loss = torch.scalar_tensor(0.0, device=graph.Z.device)
        components: dict[str, SubLossPair] = {}

        for loss in self.losses:
            loss_value = loss(model, graph, predictions)
            weighted_loss_value = loss_value * loss.weight

            total_loss += weighted_loss_value
            components[loss.name] = SubLossPair(loss_value, weighted_loss_value)

        return TotalLossResult(total_loss, components)

    # add type hints to appease mypy
    def __call__(
        self,
        model: GraphPESModel,
        graph: AtomicGraph,
        predictions: dict[PropertyKey, torch.Tensor],
    ) -> TotalLossResult:
        return super().__call__(model, graph, predictions)

    def __repr__(self) -> str:
        return "\n".join(
            ["TotalLoss:"]
            + ["    ".join(str(loss).split("\n")) for loss in self.losses]
        )


############################# CUSTOM LOSSES #############################


class PerAtomEnergyLoss(PropertyLoss):
    r"""
    A loss function that evaluates some metric on the total energy normalised
    by the number of atoms in the structure.

    .. math::
        \mathcal{L} = \text{metric}\left(
        \bigoplus_i \frac{\hat{E}_i}{N_i}, \bigoplus_i\frac{E_i}{N_i} \right)

    where :math:`\hat{E}_i` is the predicted energy for structure :math:`i`,
    :math:`E_i` is the true energy for structure :math:`i`, :math:`N_i`
    is the number of atoms in structure :math:`i` and :math:`\bigoplus_i`
    denotes the cocatenation over all structures in the batch.

    Parameters
    ----------
    metric
        The loss metric to use. Defaults to :class:`RMSE`.
    """

    def __init__(
        self,
        metric: Metric | MetricName = "RMSE",
        weight: float = 1.0,
    ):
        super().__init__("energy", metric, weight)

    def forward(
        self,
        model: GraphPESModel,
        graph: AtomicGraph,
        predictions: dict[PropertyKey, torch.Tensor],
    ) -> torch.Tensor:
        return self.metric(
            divide_per_atom(predictions["energy"], graph),
            divide_per_atom(graph.properties["energy"], graph),
        )

    @property
    def name(self) -> str:
        return f"per_atom_energy_{_get_metric_name(self.metric)}"


class ForceRMSE(PropertyLoss):
    """
    Alias for :class:`PropertyLoss` with ``property="forces"`` and
    ``metric=RMSE``.
    """

    def __init__(self, weight: float = 1.0):
        super().__init__("forces", RMSE(), weight)


## METRICS ##


def parse_metric(metric: Metric | MetricName | None) -> Metric:
    if isinstance(metric, str):
        return {"MAE": MAE(), "RMSE": RMSE(), "MSE": MSE()}[metric]

    if metric is None:
        return RMSE()

    return metric


class MSE(torch.nn.MSELoss):
    r"""
    Mean squared error metric:

    .. math::
        \frac{1}{N} \sum_i^N \left( \hat{P}_i - P_i \right)^2
    """


class RMSE(torch.nn.MSELoss):
    r"""
    Root mean squared error metric:

    .. math::
        \sqrt{ \frac{1}{N} \sum_i^N \left( \hat{P}_i - P_i \right)^2 }
    """

    def forward(
        self, input: torch.Tensor, target: torch.Tensor
    ) -> torch.Tensor:
        return (super().forward(input, target)).sqrt()


class MAE(torch.nn.L1Loss):
    r"""
    Mean absolute error metric:

    .. math::
        \frac{1}{N} \sum_i^N \left| \hat{P}_i - P_i \right|
    """


def _get_metric_name(metric: Metric) -> str:
    # if metric is a function, we want the function's name, otherwise
    # we want the metric's class name, all lowercased
    # and without the word "loss" in it

    return (
        getattr(
            metric,
            "__name__",
            metric.__class__.__name__,
        )
        .lower()
        .replace("loss", "")
    )



from __future__ import annotations

import torch
from sklearn.linear_model import Ridge

from graph_pes.atomic_graph import (
    AtomicGraph,
    number_of_structures,
    sum_per_structure,
)

from .logger import logger


def guess_per_element_mean_and_var(
    per_structure_quantity: torch.Tensor,
    batch: AtomicGraph,
    min_variance: float = 0.01,
) -> tuple[dict[int, float], dict[int, float]]:
    r"""
    Guess the per-element mean (:math:`\mu_Z`) and variance (:math:`\sigma_Z^2`)
    of a per-structure quantity using ridge regression under the following assumptions:

    1. the per-structure property, :math:`P`, is a summation over local
       properties of its components atoms: :math:`P = \sum_{i=1}^{N} p_{Z_i}`.
    2. the per-atom properties, :math:`p_{Z_i}`, are independent and identically
       distributed (i.i.d.) for each atom of type :math:`Z_i` according to a
       normal distribution: :math:`p_{Z_i} \sim \mathcal{N}(\mu_{Z_i}, \sigma_{Z_i}^2)`.

    Parameters
    ----------
    per_structure_quantity
        The per-structure quantity to guess the per-element mean and variance of.
    batch
        The batch of graphs to use for guessing the per-element mean and variance.

    Returns
    -------
    means
        A dictionary mapping atomic numbers to per-element means.
    variances
        A dictionary mapping atomic numbers to per-element variances.
    """  # noqa: E501

    # extract the atomic numbers to tensor N such that:
    # N[structure, Z] is the number of atoms of atomic number Z in structure
    unique_Zs = torch.unique(batch.Z)  # (n_Z,)
    N = torch.zeros(number_of_structures(batch), len(unique_Zs))  # (batch, n_Z)
    for i, Z in enumerate(unique_Zs):
        N[:, i] = sum_per_structure((batch.Z == Z).float(), batch)

    # calculate the per-element mean
    # use Ridge rather than LinearRegression to avoid singular matrices
    # when e.g. only one structure contains an atom of a given type...
    ridge = Ridge(fit_intercept=False, alpha=0.00001)
    ridge.fit(N.numpy(), per_structure_quantity)
    mu_Z = torch.tensor(ridge.coef_)
    means = {int(Z): float(mu) for Z, mu in zip(unique_Zs, mu_Z)}

    # calculate the per-element variance
    residuals = per_structure_quantity - N @ mu_Z
    # assuming that the squared residuals are a sum of the independent
    # variances for each atom, we can estimate these variances again
    # using Ridge regression
    ridge.fit(N.numpy(), residuals**2)
    var_Z = ridge.coef_
    # avoid negative variances by clipping to min value
    variances = {
        int(Z): max(float(var), min_variance)
        for Z, var in zip(unique_Zs, var_Z)
    }

    logger.debug(f"Per-element means: {means}")
    logger.debug(f"Per-element variances: {variances}")

    return means, variances



from __future__ import annotations

import warnings

import torch
from ase.data import atomic_numbers

from graph_pes.atomic_graph import AtomicGraph, PropertyKey
from graph_pes.graph_pes_model import GraphPESModel
from graph_pes.utils.logger import logger
from graph_pes.utils.nn import PerElementParameter
from graph_pes.utils.shift_and_scale import guess_per_element_mean_and_var


class EnergyOffset(GraphPESModel):
    r"""
    A model that predicts the total energy as a sum of per-species offsets:

    .. math::
        E(\mathcal{G}) = \sum_i \varepsilon_{Z_i}

    where :math:`\varepsilon_{Z_i}` is the energy offset for atomic species
    :math:`Z_i`.

    With our chemistry hat on, this is equivalent to a PerfectGas model:
    there is no contribution to the total energy from the interactions between
    atoms. Hence, this model generates zero for both force and stress
    predictions.

    Parameters
    ----------
    offsets
        The energy offsets for each atomic species.
    """

    def __init__(self, offsets: PerElementParameter):
        super().__init__(
            cutoff=0,
            implemented_properties=["local_energies"],
        )
        self._offsets = offsets

    def forward(self, graph: AtomicGraph) -> dict[PropertyKey, torch.Tensor]:
        return {
            "local_energies": self._offsets[graph.Z].squeeze(),
        }

    def non_decayable_parameters(self) -> list[torch.nn.Parameter]:
        """The ``_offsets`` parameter should not be decayed."""
        return [self._offsets]

    def __repr__(self):
        return self._offsets._repr(alias=self.__class__.__name__)


class FixedOffset(EnergyOffset):
    """
    An :class:`~graph_pes.models.offsets.EnergyOffset` model with pre-defined
    and fixed energy offsets for each element. These do not change during
    training. Any element not specified in the ``final_values`` argument will
    be assigned an energy offset of zero.

    Parameters
    ----------
    final_values
        A dictionary mapping element symbols to fixed energy offset values.

    Examples
    --------
    >>> model = FixedOffset(H=-1.3, C=-13.0)
    """

    def __init__(self, **final_values: float):
        offsets = PerElementParameter.from_dict(
            **final_values,
            requires_grad=False,
        )
        super().__init__(offsets)


class LearnableOffset(EnergyOffset):
    """
    An :class:`~graph_pes.models.offsets.EnergyOffset` model with
    learnable energy offsets for each element.

    During pre-fitting, for each element in the training data not specified by
    the user, the model will estimate the energy offset from the data using
    ridge regression (see
    :func:`~graph_pes.utils.shift_and_scale.guess_per_element_mean_and_var`).

    Parameters
    ----------
    initial_values
        A dictionary of initial energy offsets for each atomic species.
        Leave this empty to guess the offsets from the training data.

    Examples
    --------
    Estimate all relevant energy offsets from the training data:

    >>> model = LearnableOffset()
    >>> # estimates offsets from data
    >>> model.pre_fit_all_components(training_data)

    Specify some initial values for the energy offsets:

    >>> model = LearnableOffset(H=0.0, C=-3.0)
    >>> # estimate offsets for elements that aren't C or H
    >>> model.pre_fit_all_components(training_data)
    """

    def __init__(self, **initial_values: float):
        offsets = PerElementParameter.from_dict(
            **initial_values,
            requires_grad=True,
        )
        super().__init__(offsets)

        Zs = torch.tensor(
            [atomic_numbers[symbol] for symbol in initial_values],
            dtype=torch.long,
        )
        self.register_buffer("_pre_specified_Zs", Zs)

    @torch.no_grad()
    def pre_fit(self, graphs: AtomicGraph) -> None:
        """
        Calculate the **mean** energy offsets per element from the training data
        using linear regression.

        **Note**: this will lead to a lack of physicality in the model, since
        there is now no guarantee that the energy of an isolated atom, or the
        dissociation limit, is correctly modelled.

        Parameters
        ----------
        graphs
            The training data.
        """

        if "energy" not in graphs.properties:
            warnings.warn(
                "No energy labels found in the training data. "
                "Can't guess suitable per-element energy offsets for "
                f"{self.__class__.__name__}.",
                stacklevel=2,
            )
            return

        # use ridge regression to estimate the mean energy contribution
        # from each atomic species
        offsets, _ = guess_per_element_mean_and_var(
            graphs.properties["energy"], graphs
        )
        for z, offset in offsets.items():
            if torch.any(self._pre_specified_Zs == z):
                continue
            self._offsets[z] = offset

        logger.warning(f"""
Estimated per-element energy offsets from the training data:
    {self._offsets}
This will lead to a lack of guaranteed physicality in the model, 
since the energy of an isolated atom (and hence the behaviour of 
this model in the dissociation limit) is not guaranteed to be 
correct. Use a FixedOffset energy offset model if you require
and know the reference energy offsets.""")



from __future__ import annotations

import copy
import time
from abc import ABC
from pathlib import Path
from typing import Any, Literal, Mapping, cast

import pytorch_lightning as pl
import torch
from ase.data import chemical_symbols
from pytorch_lightning import Callback, LightningModule, Trainer
from pytorch_lightning.callbacks import ProgressBar, StochasticWeightAveraging
from pytorch_lightning.callbacks.early_stopping import EarlyStopping
from pytorch_lightning.loggers import Logger
from pytorch_lightning.loggers import WandbLogger as PTLWandbLogger
from typing_extensions import override

from graph_pes.graph_pes_model import GraphPESModel
from graph_pes.models.addition import AdditionModel
from graph_pes.models.components.scaling import LocalEnergiesScaler
from graph_pes.models.offsets import LearnableOffset
from graph_pes.training.utils import VALIDATION_LOSS_KEY
from graph_pes.utils.lammps import deploy_model
from graph_pes.utils.logger import logger
from graph_pes.utils.misc import uniform_repr


class GraphPESCallback(Callback, ABC):
    """
    A base class for all callbacks that require access to useful
    information generated by the ``graph-pes-train`` command.
    """

    def __init__(self):
        self.root: Path = None  # type: ignore

    def _register_root(self, root: Path):
        # called by us before any training starts
        self.root = root

    def get_model(self, pl_module: LightningModule) -> GraphPESModel:
        return cast(GraphPESModel, pl_module.model)

    def get_model_on_cpu(self, pl_module: LightningModule) -> GraphPESModel:
        model = self.get_model(pl_module)
        model = copy.deepcopy(model)
        model.to("cpu")
        return model

    def __repr__(self) -> str:
        root = self.root if hasattr(self, "root") else None
        return f"{self.__class__.__name__}(root={root})"


class DumpModel(GraphPESCallback):
    """
    Dump the model to ``<output_dir>/dumps/model_{epoch}.pt``
    at regular intervals.

    Parameters
    ----------
    every_n_val_checks: int
        The number of validation epochs between dumps.
    """

    def __init__(self, every_n_val_checks: int = 10):
        self.every_n_val_checks = every_n_val_checks

    def on_validation_epoch_end(
        self, trainer: Trainer, pl_module: LightningModule
    ):
        if not trainer.is_global_zero:
            return

        epoch = trainer.current_epoch
        if epoch % self.every_n_val_checks != 0:
            return

        model_path = self.root / "dumps" / f"model_{epoch}.pt"
        model_path.parent.mkdir(exist_ok=True)
        torch.save(self.get_model_on_cpu(pl_module), model_path)


def log_offset(model: GraphPESModel, logger: Logger):
    if not isinstance(model, AdditionModel):
        return

    offsets = [
        c for c in model.models.values() if isinstance(c, LearnableOffset)
    ]
    if not offsets:
        return

    Zs = offsets[0]._offsets._accessed_Zs
    logger.log_metrics(
        {
            f"offset/{chemical_symbols[Z]}": offsets[0]._offsets[Z].item()
            for Z in Zs
        }
    )


class OffsetLogger(GraphPESCallback):
    """
    Log any learned, per-element offsets of the model at the
    end of each validation epoch.
    """

    def on_validation_epoch_end(
        self, trainer: Trainer, pl_module: LightningModule
    ):
        if not trainer.is_global_zero:
            return

        if not trainer.logger:
            return

        log_offset(self.get_model(pl_module), trainer.logger)


def log_scales(model: GraphPESModel, logger: Logger, name: str | None = None):
    if isinstance(model, AdditionModel):
        for name, child in model.models.items():
            log_scales(child, logger, name)
        return

    prefix = f"scale/{name}" if name else "scale"
    scaler = next(
        (c for c in model.modules() if isinstance(c, LocalEnergiesScaler)), None
    )
    if not scaler:
        return
    scaling = scaler.per_element_scaling
    logger.log_metrics(
        {
            f"{prefix}/{chemical_symbols[Z]}": scaling[Z].item()
            for Z in scaling._accessed_Zs
        }
    )


class ScalesLogger(GraphPESCallback):
    """
    Log any learned, per-element scaling factors of the model at the
    end of each validation epoch.
    """

    def on_validation_epoch_end(
        self, trainer: Trainer, pl_module: LightningModule
    ):
        if not trainer.is_global_zero:
            return

        if not trainer.logger:
            return

        log_scales(self.get_model(pl_module), trainer.logger)


class EarlyStoppingWithLogging(EarlyStopping, GraphPESCallback):
    """
    Log various information relating to the early stopping process:

    * number of validation checks since the best validation loss was observed
    * "distances above" the best validation loss
    """

    def __init__(
        self,
        monitor: str,
        patience: int,
        min_delta: float = 0.0,
        **kwargs,
    ):
        super().__init__(monitor, min_delta, patience, **kwargs)
        self._min_delta = min_delta
        self.state = {
            "best_val_loss": float("inf"),
            "best_val_loss_check": 0,
            "total_checks": 0,
        }

    def on_validation_epoch_end(
        self, trainer: Trainer, pl_module: LightningModule
    ):
        super().on_validation_epoch_end(trainer, pl_module)

        if not trainer.is_global_zero or not trainer.logger:
            return

        self.state["total_checks"] += 1
        current_loss = trainer.callback_metrics.get(VALIDATION_LOSS_KEY, None)
        if current_loss is None:
            return
        current_loss = current_loss.item()

        if current_loss < self.state["best_val_loss"] - self._min_delta:
            self.state["best_val_loss"] = current_loss
            self.state["best_val_loss_check"] = self.state["total_checks"]

        checks_since_best = (
            self.state["total_checks"] - self.state["best_val_loss_check"]
        )
        distance_above_best = current_loss - self.state["best_val_loss"]

        trainer.logger.log_metrics(
            {
                "early_stopping/checks_since_best": checks_since_best,
                "early_stopping/best_valid_loss": self.state["best_val_loss"],
                "early_stopping/distance_above_best": distance_above_best,
            }
        )

    def load_state_dict(self, state_dict: dict):
        self.state.update(state_dict)

    def state_dict(self) -> dict:
        return self.state


class SaveBestModel(GraphPESCallback):
    """
    Save the best model to ``<output_dir>/model.pt`` and deploy it to
    ``<output_dir>/lammps_model.pt``.
    """

    def __init__(self):
        super().__init__()
        self.best_val_loss = float("inf")
        self.try_to_deploy = True

    def on_validation_end(self, trainer: Trainer, pl_module: LightningModule):
        if not trainer.is_global_zero or not self.root:
            return

        # get current validation loss
        current_loss = trainer.callback_metrics.get(VALIDATION_LOSS_KEY, None)
        if current_loss is None:
            return

        # if the validation loss has improved, save the model
        if current_loss < self.best_val_loss:
            self.best_val_loss = current_loss
            cpu_model = self.get_model_on_cpu(pl_module)
            logger.debug(f"New best model: validation loss {current_loss}")

            # log the final path to the trainer.logger.summary
            model_path = self.root / "model.pt"
            lammps_model_path = self.root / "lammps_model.pt"

            assert trainer.logger is not None
            trainer.logger.log_hyperparams(
                {
                    "model_path": model_path,
                    "lammps_model_path": lammps_model_path,
                }
            )
            torch.save(cpu_model, model_path)
            logger.debug(f"Model saved to {model_path}")

            if self.try_to_deploy:
                try:
                    deploy_model(cpu_model, path=lammps_model_path)
                    logger.debug(
                        f"Deployed model for use with LAMMPS to "
                        f"{lammps_model_path}"
                    )
                except Exception as e:
                    logger.warning(
                        f"Failed to deploy model for use with LAMMPS: {e}"
                    )
                    self.try_to_deploy = False


class ModelTimer(pl.Callback):
    def __init__(self):
        super().__init__()
        self.tick_ms: float | None = None

    def start(self):
        self.tick_ms = time.time_ns() // 1_000_000

    def stop(
        self, pl_module: pl.LightningModule, stage: Literal["train", "valid"]
    ):
        assert self.tick_ms is not None
        duration_ms = max((time.time_ns() // 1_000_000) - self.tick_ms, 1)
        self.tick_ms = None

        for name, x in (
            ("step_duration_ms", duration_ms),
            ("its_per_s", 1_000 / duration_ms),
        ):
            pl_module.log(
                f"timer/{name}/{stage}",
                x,
                batch_size=1,
                on_epoch=stage == "valid",
                on_step=stage == "train",
                prog_bar=name == "its_per_s",
                sync_dist=stage == "valid",
            )

    @override
    def on_train_batch_start(self, *args, **kwargs):
        self.start()

    @override
    def on_train_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        *args,
        **kwargs,
    ):
        self.stop(pl_module, "train")

    @override
    def on_validation_batch_start(self, *args, **kwargs):
        self.start()

    @override
    def on_validation_batch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        *args,
        **kwargs,
    ):
        self.stop(pl_module, "valid")


class VerboseSWACallback(StochasticWeightAveraging):
    @override
    def on_train_epoch_start(
        self, trainer: pl.Trainer, pl_module: pl.LightningModule
    ) -> None:
        if (not self._initialized) and (
            self.swa_start <= trainer.current_epoch <= self.swa_end
        ):
            logger.info("SWA: starting SWA")

        return super().on_train_epoch_start(trainer, pl_module)


class LoggedProgressBar(ProgressBar):
    """
    A progress bar that logs all metrics at the end of each validation epoch.
    """

    def __init__(self):
        super().__init__()
        self._enabled = True
        self._start_time = time.time()
        self._widths: dict[str, int] = {}

    @override
    def on_train_start(
        self, trainer: pl.Trainer, pl_module: pl.LightningModule
    ):
        self._start_time = time.time()

    @override
    def disable(self):
        self._enabled = False

    @override
    def enable(self):
        self._enabled = True

    @override
    def on_validation_epoch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
    ):
        if not self._enabled or trainer.sanity_checking:
            return

        metrics = self.get_metrics(trainer, pl_module)

        if not self._widths:
            # first time we're logging things:
            # calculate the widths of the columns, and print the headers

            # Split headers with more than 1 slash
            # over two rows
            def split_header(h: str) -> list[str]:
                if "/" in h:
                    parts = h.split("/")
                    if len(parts) > 2:
                        return [f"{parts[0]}/{parts[1]}", "/".join(parts[2:])]
                return [h]

            headers = list(metrics.keys())
            split_headers = [split_header(h) for h in headers]

            header_widths = {
                header: max(len(line) for line in lines)
                for header, lines in zip(headers, split_headers)
            }
            content_widths = {
                header: len(metrics[header]) for header in headers
            }
            self._widths = {
                header: max(header_widths[header], content_widths[header])
                + (6 if header == "time" else 3)
                for header in headers
            }

            # print the headers
            first_row = [
                "" if len(lines) == 1 else lines[0] for lines in split_headers
            ]
            second_row = [lines[-1] for lines in split_headers]

            print(
                "".join(
                    f"{part:>{self._widths[header]}}"
                    for part, header in zip(first_row, headers)
                ),
                flush=True,
            )
            print(
                "".join(
                    f"{part:>{self._widths[header]}}"
                    for part, header in zip(second_row, headers)
                ),
                flush=True,
            )

        # print the values for this epoch
        print(
            "".join(f"{v:>{self._widths[k]}}" for k, v in metrics.items()),
            flush=True,
        )

    @override
    def get_metrics(  # type: ignore
        self, trainer: pl.Trainer, pl_module: pl.LightningModule
    ) -> dict[str, str]:
        def logged_value(v: float | int | Any):
            return f"{v:.5f}" if isinstance(v, float) else str(v)

        metrics = {"epoch": f"{trainer.current_epoch + 1:>5}"}
        super_metrics = super().get_metrics(trainer, pl_module)
        super_metrics.pop("v_num", None)
        for k, v in super_metrics.items():
            metrics[k] = logged_value(v)

        # rearrange according to: epoch | time | valid/* | rest...
        sorted_metrics = {"epoch": metrics.pop("epoch")}
        sorted_metrics["time"] = f"{time.time() - self._start_time:.1f}"
        for k in list(metrics):
            if k.startswith("valid/"):
                sorted_metrics[k] = metrics.pop(k)
        sorted_metrics.update(metrics)

        return sorted_metrics


class WandbLogger(PTLWandbLogger):
    """A subclass of WandbLogger that automatically sets the id and save_dir."""

    def __init__(self, output_dir: Path, log_epoch: bool, **kwargs):
        if "id" not in kwargs:
            kwargs["id"] = output_dir.name
        if "save_dir" not in kwargs:
            kwargs["save_dir"] = str(output_dir.parent)
        super().__init__(**kwargs)
        self._kwargs = kwargs
        self._log_epoch = log_epoch

    def log_metrics(
        self, metrics: Mapping[str, float], step: int | None = None
    ):
        if not self._log_epoch:
            metrics = {k: v for k, v in metrics.items() if k != "epoch"}
        return super().log_metrics(metrics, step)

    def __repr__(self):
        return uniform_repr(self.__class__.__name__, **self._kwargs)



from __future__ import annotations

from typing import Any, Union, cast

import e3nn.util.jit
import torch
import torch.fx
from e3nn import o3


class LinearReadOut(o3.Linear):
    """
    Map a set of features with arbitrary irreps to a single irrep with
    a single feature using a linear layer.

    Parameters
    ----------
    input_irreps : str or o3.Irreps
        The irreps of the input features.
    output_irrep : str, optional
        The irrep of the output feature. Defaults to "0e".

    Examples
    --------

    Map an embedding to a scalar output:

    >>> LinearReadOut("16x0e+16x1o+16x2e")
    LinearReadOut(16x0e+16x1o+16x2e -> 1x0e | 16 weights)


    Map an embedding to a vector output:

    >>> LinearReadOut("16x0e+16x1o+16x2e", "1o")
    LinearReadOut(16x0e+16x1o+16x2e -> 1x1o | 16 weights)
    """

    def __init__(self, input_irreps: str | o3.Irreps, output_irrep: str = "0e"):
        super().__init__(input_irreps, f"1x{output_irrep}")

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        return super().__call__(x)


def _get_activation(name: str) -> torch.nn.Module:
    try:
        return getattr(torch.nn, name)()
    except AttributeError:
        raise ValueError(f"Unknown activation function: {name}") from None


class NonLinearReadOut(torch.nn.Sequential):
    """
    Non-linear readout layer for equivariant neural networks.

    This class implements a non-linear readout layer that takes input features
    with arbitrary irreps and produces a scalar output. It uses a linear layer
    followed by an activation function and another linear layer to produce
    the final scalar output.

    Parameters
    ----------
    input_irreps : str
        The irreps of the input features.
    output_irrep : str, optional
        The irrep of the output feature. Defaults to "0e".
    hidden_dim : int, optional
        The dimension of the hidden layer. If None,
        it defaults to the number of scalar output irreps in the input.
    activation : str or torch.nn.Module, optional
        The activation function to use. Can be specified as a string
        (e.g., 'ReLU', 'SiLU') or as a torch.nn.Module. Defaults to ``SiLU``
        for even output irreps and ``Tanh`` for odd output irreps.
        **Care must be taken to ensure that the activation is suitable for
        the target irreps!**

    Examples
    --------

    Map an embedding to a scalar output:

    >>> NonLinearReadOut("16x0e+16x1o+16x2e")
    NonLinearReadOut(
      (0): Linear(16x0e+16x1o+16x2e -> 16x0e | 256 weights)
      (1): SiLU()
      (2): Linear(16x0e -> 1x0e | 16 weights)
    )

    Map an embedding to a vector output:

    >>> NonLinearReadOut("16x0e+16x1o+16x2e", "1o")
    NonLinearReadOut(
      (0): Linear(16x0e+16x1o+16x2e -> 16x1o | 256 weights)
      (1): Tanh()
      (2): Linear(16x1o -> 1x1o | 16 weights)
    )
    """

    def __init__(
        self,
        input_irreps: str | o3.Irreps,
        output_irrep: str = "0e",
        hidden_dim: int | None = None,
        activation: str | torch.nn.Module | None = None,
    ):
        if activation is None:
            activation = "SiLU" if "e" in str(output_irrep) else "Tanh"

        hidden_dim = (
            o3.Irreps(input_irreps).count(o3.Irrep(output_irrep))
            if hidden_dim is None
            else hidden_dim
        )

        if isinstance(activation, str):
            activation = _get_activation(activation)
        elif not isinstance(activation, torch.nn.Module):
            raise ValueError("activation must be a string or a torch.nn.Module")

        super().__init__(
            o3.Linear(input_irreps, f"{hidden_dim}x{output_irrep}"),
            activation,
            o3.Linear(f"{hidden_dim}x{output_irrep}", f"1x{output_irrep}"),
        )

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        return self.forward(x)


ReadOut = Union[LinearReadOut, NonLinearReadOut]


@e3nn.util.jit.compile_mode("script")
class SphericalHarmonics(o3.SphericalHarmonics):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __repr__(self):
        return f"SphericalHarmonics(1x1o -> {self.irreps_out})"


def build_limited_tensor_product(
    node_embedding_irreps: o3.Irreps,
    edge_embedding_irreps: o3.Irreps,
    allowed_outputs: list[o3.Irrep],
) -> o3.TensorProduct:
    # we want to build a tensor product that takes the:
    # - node embeddings of each neighbour (node_irreps_in)
    # - spherical-harmonic expansion of the neighbour directions
    #      (o3.Irreps.spherical_harmonics(l_max) = e.g. 1x0e, 1x1o, 1x2e)
    # and generates
    # - message embeddings from each neighbour (node_irreps_out)
    #
    # crucially, rather than using the full tensor product, we limit the
    # output irreps to be of order l_max at most. we do this by defining a
    # sequence of instructions that specify the connectivity between the
    # two input irreps and the output irreps
    #
    # we build this instruction set by naively iterating over all possible
    # combinations of input irreps and spherical-harmonic irreps, and
    # filtering out those that are above the desired order
    #
    # finally, we sort the instructions so that the tensor product generates
    # a tensor where all elements of the same irrep are grouped together
    # this aids normalisation in subsequent operations

    output_irreps = []
    instructions = []

    for i, (channels, ir_in) in enumerate(node_embedding_irreps):
        # the spherical harmonic expansions always have 1 channel per irrep,
        # so we don't care about their channel dimension
        for l, (_, ir_edge) in enumerate(edge_embedding_irreps):
            # get all possible output irreps that this interaction could
            # generate, e.g. 1e x 1e -> 0e + 1e + 2e
            possible_output_irreps = ir_in * ir_edge

            for ir_out in possible_output_irreps:
                # (order, parity) = ir_out
                if ir_out not in allowed_outputs:
                    continue

                # if we want this output from the tensor product, add it to the
                # list of instructions
                k = len(output_irreps)
                output_irreps.append((channels, ir_out))
                # from the i'th irrep of the neighbour embedding
                # and from the l'th irrep of the spherical harmonics
                # to the k'th irrep of the output tensor
                instructions.append((i, l, k, "uvu", True))

    # since many paths can lead to the same output irrep, we sort the
    # instructions so that the tensor product generates tensors in a
    # nice order, e.g. 32x0e + 16x1o, not 16x0e + 16x1o + 16x0e
    output_irreps = o3.Irreps(output_irreps)
    assert isinstance(output_irreps, o3.Irreps)
    output_irreps, permutation, _ = output_irreps.sort()

    # permute the output indexes of the instructions to match the sorted irreps:
    instructions = [
        (i_in1, i_in2, permutation[i_out], mode, train)
        for i_in1, i_in2, i_out, mode, train in instructions
    ]

    return o3.TensorProduct(
        node_embedding_irreps,
        edge_embedding_irreps,
        output_irreps,
        instructions,
        # this tensor product will be parameterised by weights that are learned
        # from neighbour distances, so it has no internal weights
        internal_weights=False,
        shared_weights=False,
    )


def as_irreps(input: Any) -> o3.Irreps:
    # util to precent checking isinstance(o3.Irreps) all the time
    return cast(o3.Irreps, o3.Irreps(input))


def to_full_irreps(n_features: int, irreps: list[o3.Irrep]) -> o3.Irreps:
    # convert a list of irreps to a full irreps object
    return as_irreps([(n_features, ir) for ir in irreps])



from __future__ import annotations

import argparse
import ast
import contextlib
import csv
import os
import random
from pathlib import Path

import numpy as np
import torch
import yaml
from pytorch_lightning.loggers import CSVLogger, Logger

from graph_pes.config.shared import TorchConfig
from graph_pes.training.callbacks import WandbLogger
from graph_pes.utils import distributed
from graph_pes.utils.logger import logger
from graph_pes.utils.misc import (
    build_single_nested_dict,
    nested_merge,
    nested_merge_all,
)


def extract_config_dict_from_command_line(description: str) -> dict:
    parser = argparse.ArgumentParser(
        description=description,
        epilog="Copyright 2023-24, John Gardner",
    )
    parser.add_argument(
        "args",
        nargs="*",
        help=(
            "Config files and command line specifications. "
            "Config files should be YAML (.yaml/.yml) files. "
            "Command line specifications should be in the form "
            "my/nested/key=value. "
            "Final config is built up from these items in a left "
            "to right manner, with later items taking precedence "
            "over earlier ones in the case of conflicts. "
        ),
    )
    args = parser.parse_args()
    return nested_merge_all(*map(get_data_from_cli_arg, args.args))


def get_data_from_cli_arg(arg: str) -> dict:
    if arg.endswith(".yaml") or arg.endswith(".yml"):
        # it's a config file
        try:
            with open(arg) as f:
                return yaml.safe_load(f)
        except Exception as e:
            logger.error(
                f"You specified a config file ({arg}) " "that we couldn't load."
            )
            raise e

    elif "=" in arg:
        # it's an override
        key, value = arg.split("=", maxsplit=1)
        keys = key.split("/")

        # parse the value
        with contextlib.suppress(yaml.YAMLError):
            value = yaml.safe_load(value)

        return build_single_nested_dict(keys, value)

    raise ValueError(
        f"Invalid argument: {arg}. "
        "Expected a YAML file or an override in the form key=value"
    )


def configure_general_options(torch_config: TorchConfig, seed: int):
    prec = torch_config.float32_matmul_precision
    torch.set_float32_matmul_precision(prec)
    logger.debug(f"Using {prec} precision for float32 matrix multiplications.")

    ftype = torch_config.dtype
    logger.debug(f"Using {ftype} as default dtype.")
    torch.set_default_dtype(
        {
            "float16": torch.float16,
            "float32": torch.float32,
            "float64": torch.float64,
        }[ftype]
    )
    # a nice setting for e3nn components that get scripted upon instantiation
    # - DYNAMIC refers to the fact that they will expect different input sizes
    #   at every iteration (graphs are not all the same size)
    # - 4 is the number of times we attempt to recompile before giving up
    torch.jit.set_fusion_strategy([("DYNAMIC", 4)])

    # a non-verbose version of pl.seed_everything
    logger.debug(f"Using seed {seed} for reproducibility.")
    os.environ["PL_GLOBAL_SEED"] = str(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    os.environ["PL_SEED_WORKERS"] = "0"


def extract_summary(logs: Logger) -> dict:
    if isinstance(logs, WandbLogger):
        return dict(logs.experiment.summary)

    elif isinstance(logs, CSVLogger):
        try:
            experiment = logs.experiment
            with experiment._fs.open(
                experiment.metrics_file_path, "r", newline=""
            ) as file:
                metrics = list(csv.DictReader(file))
        except Exception as e:
            logger.warning(f"Failed to read metrics file: {e}")
            return {}

        summary = {}
        for metric_row in metrics:
            m = {
                k: ast.literal_eval(v) for k, v in metric_row.items() if v != ""
            }
            summary.update(m)
        return summary

    else:
        logger.warning(
            f"Unsupported logger type: {type(logs)}. "
            "We can't extract a summary."
        )
        return {}


def update_summary(logger: Logger | None, summary_file: Path):
    if logger is None:
        return
    if not distributed.IS_RANK_0:
        return

    if summary_file.exists():
        with open(summary_file) as f:
            existing = yaml.safe_load(f)
    else:
        existing = {}

    summary = extract_summary(logger)
    summary_file.parent.mkdir(parents=True, exist_ok=True)
    with open(summary_file, "w") as f:
        yaml.dump(nested_merge(existing, summary), f)



import torch

from graph_pes import AtomicGraph, GraphPESModel
from graph_pes.atomic_graph import (
    PropertyKey,
    neighbour_distances,
    neighbour_vectors,
    number_of_edges,
    sum_per_structure,
)
from graph_pes.utils.threebody import (
    angle_spanned_by,
    triplet_edge_pairs,
)


class MatterSim_M3Gnet_Wrapper(GraphPESModel):
    def __init__(self, model: torch.nn.Module):
        super().__init__(
            cutoff=model.model_args["cutoff"],  # type: ignore
            implemented_properties=["local_energies"],
        )
        self.model = model
        self.threebody_cutoff = torch.tensor(
            model.model_args["threebody_cutoff"]
        )  # type: ignore

    def forward(self, graph: AtomicGraph) -> dict[PropertyKey, torch.Tensor]:
        # pre-compute
        edge_lengths = neighbour_distances(graph)  # (E)
        edge_pairs = triplet_edge_pairs(graph, self.threebody_cutoff.item())
        triplets_per_leading_edge = count_number_of_triplets_per_leading_edge(
            edge_pairs, graph
        )
        r_ik = edge_lengths[edge_pairs[:, 1]]
        v = neighbour_vectors(graph)
        v_ij = v[edge_pairs[:, 0]]
        v_ik = v[edge_pairs[:, 1]]
        angle = angle_spanned_by(v_ij, v_ik)

        num_atoms = sum_per_structure(
            torch.ones_like(graph.Z), graph
        ).unsqueeze(-1)

        # num_bonds is of shape (n_structures,) such that
        # num_bonds[i] = sum(graph.neighbour_list[0] == i)
        bonds_per_atom = torch.zeros_like(graph.Z)
        bonds_per_atom = bonds_per_atom.scatter_add(
            dim=0,
            index=graph.neighbour_list[0],
            src=torch.ones_like(graph.neighbour_list[0]),
        )
        num_bonds = sum_per_structure(bonds_per_atom, graph).unsqueeze(-1)

        three_body_indices = edge_pairs
        num_triple_ij = triplets_per_leading_edge.unsqueeze(-1)

        # use the forward pass of M3Gnet
        atom_attr = self.model.atom_embedding(self.model.one_hot_atoms(graph.Z))
        edge_attr = self.model.rbf(edge_lengths)
        edge_attr_zero = edge_attr
        edge_attr = self.model.edge_encoder(edge_attr)
        three_basis = self.model.sbf(r_ik, angle)

        for conv in self.model.graph_conv:
            atom_attr, edge_attr = conv(
                atom_attr,
                edge_attr,
                edge_attr_zero,
                graph.neighbour_list,
                three_basis,
                three_body_indices,
                edge_lengths.unsqueeze(-1),
                num_bonds,
                num_triple_ij,
                num_atoms,
            )

        local_energies = self.model.final(atom_attr).view(-1)
        local_energies = self.model.normalizer(local_energies, graph.Z)

        return {"local_energies": local_energies}


def mattersim(load_path: str = "mattersim-v1.0.0-1m") -> GraphPESModel:
    """
    Load a ``mattersim`` model from a checkpoint file, and convert it to a
    :class:`~graph_pes.GraphPESModel` on the CPU.

    Parameters
    ----------
    load_path: str
        The path to the ``mattersim`` checkpoint file. Expected to be one of
        ``mattersim-v1.0.0-1m`` or ``mattersim-v1.0.0-5m`` currently.
    """
    from mattersim.forcefield.potential import Potential

    model = Potential.from_checkpoint(  # type: ignore
        load_path,
        load_training_state=False,  # only load the model
        device="cpu",  # manage the device ourself later
    ).model
    return MatterSim_M3Gnet_Wrapper(model)


@torch.no_grad()
def count_number_of_triplets_per_leading_edge(
    edge_pairs: torch.Tensor,
    graph: AtomicGraph,
):
    """
    Return ``T`` of shape ``(E,)`` where ``T[e]`` is the number of edge pairs
    that have edge number ``e`` as the first edge in the pair.

    Parameters
    ----------
    edge_pairs: torch.Tensor
        A ``(E, 2)`` shaped tensor indicating pairs of edges that form a
        triplet ``(i, j, k)`` (see :func:`triplet_edge_pairs`).
    graph: AtomicGraph
        The graph from which the edge pairs were derived.

    Returns
    -------
    triplets_per_edge: torch.Tensor
        A ``(E,)`` shaped tensor where ``triplets_per_edge[e]`` is the
        number of edge pairs that have edge ``e`` as the first edge in the
        pair.

    """
    triplets_per_edge = torch.zeros(
        number_of_edges(graph), device=graph.R.device, dtype=torch.long
    )
    return triplets_per_edge.scatter_add(
        dim=0,
        index=edge_pairs[:, 0],
        src=torch.ones_like(edge_pairs[:, 0]),
    )



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Sequence, cast

import opt_einsum_fx
import torch
import torch.fx
from e3nn import o3
from e3nn.util.codegen import CodeGenMixin

from graph_pes.models.e3nn.utils import to_full_irreps
from graph_pes.utils.nn import UniformModuleList

BATCH_DIM_EG = 20
SPARE = "wxvnzrtyuops"


def _get_optimised_summation(
    _lambda: Callable,
    example_input_sizes: list[Sequence[int]],
) -> torch.fx.GraphModule:
    inputs = tuple(torch.randn(size) for size in example_input_sizes)
    opt = opt_einsum_fx.optimize_einsums_full(
        model=torch.fx.symbolic_trace(_lambda), example_inputs=inputs
    )
    return cast(torch.fx.GraphModule, opt)


@dataclass
class ContractionConfig:
    num_features: int
    irrep_s_in: list[o3.Irrep]
    irrep_out: o3.Irrep
    n_node_attributes: int

    def __repr__(self) -> str:
        _in = to_full_irreps(self.num_features, self.irrep_s_in)
        _out = to_full_irreps(self.num_features, [self.irrep_out])
        return f"{_in} + {self.n_node_attributes}x0e -> {_out}"


class InitialContraction(CodeGenMixin, torch.nn.Module):
    def __init__(self, config: ContractionConfig, correlation: int):
        super().__init__()

        # U is of shape (X, (Y,) * correlation, Z)
        # where X = irrep_out.dim
        self.register_buffer(
            "U", get_U_matrix(config.irrep_s_in, config.irrep_out, correlation)
        )
        Y = self.U.size()[-2]
        Z = self.U.size()[-1]

        self.W = torch.nn.Parameter(
            torch.randn(config.n_node_attributes, Z, config.num_features) / Z
        )

        # the contraction is a summation that takes 4 inputs:
        # U, W, node_embeddings, node_attributes
        instruction = "ik,ekc,bci,be -> bc"
        # we parallelise over the "spare" dimensions in U
        spare_dims = SPARE[:correlation]
        instruction = f"{spare_dims}{instruction}{spare_dims}"

        self.summation = _get_optimised_summation(
            lambda x, y, z, w: torch.einsum(instruction, x, y, z, w),
            [
                (self.U.shape),
                (config.n_node_attributes, Z, config.num_features),
                (BATCH_DIM_EG, config.num_features, Y),
                (BATCH_DIM_EG, config.n_node_attributes),
            ],
        )

    def forward(
        self,
        node_embeddings: torch.Tensor,
        node_attributes: torch.Tensor,
    ) -> torch.Tensor:
        return self.summation(self.U, self.W, node_embeddings, node_attributes)

    def __call__(
        self,
        node_embeddings: torch.Tensor,
        node_attributes: torch.Tensor,
    ) -> torch.Tensor:
        return super().__call__(node_embeddings, node_attributes)


class FollowingWeightContraction(CodeGenMixin, torch.nn.Module):
    def __init__(self, config: ContractionConfig, correlation: int):
        super().__init__()

        # as above, U is of shape (X, (Y,) * correlation, Z)
        self.register_buffer(
            "U", get_U_matrix(config.irrep_s_in, config.irrep_out, correlation)
        )
        Z = self.U.size()[-1]

        self.W = torch.nn.Parameter(
            torch.randn(config.n_node_attributes, Z, config.num_features) / Z
        )

        # this contraction acts on U, W and the node attributes
        # spare_dims = get_spare_dims(correlation, irrep_out)
        spare_dims = SPARE[: correlation + 1]
        instruction = f"{spare_dims}k,ekc,be->bc{spare_dims}"

        self.summation = _get_optimised_summation(
            lambda x, y, z: torch.einsum(instruction, x, y, z),
            [
                (self.U.shape),
                (config.n_node_attributes, Z, config.num_features),
                (BATCH_DIM_EG, config.n_node_attributes),
            ],
        )

    def forward(
        self,
        node_attributes: torch.Tensor,
    ) -> torch.Tensor:
        return self.summation(self.U, self.W, node_attributes)

    # for mypy
    def __call__(
        self,
        node_attributes: torch.Tensor,
    ) -> torch.Tensor:
        return super().__call__(node_attributes)


class FeatureContraction(CodeGenMixin, torch.nn.Module):
    def __init__(self, config: ContractionConfig, correlation: int):
        super().__init__()

        # as above, U is of shape (X, (Y,) * correlation, Z)
        U = get_U_matrix(config.irrep_s_in, config.irrep_out, correlation)
        X = config.irrep_out.dim
        Y = U.size()[-2]

        # this contraction acts on the output of a weight contraction
        # and the node features
        spare_dims = SPARE[:correlation]
        instruction = f"bc{spare_dims}i,bci->bc{spare_dims}"

        self.summation = _get_optimised_summation(
            lambda x, y: torch.einsum(instruction, x, y),
            [
                [BATCH_DIM_EG, config.num_features, X] + [Y] * correlation,
                (BATCH_DIM_EG, config.num_features, Y),
            ],
        )

    def forward(
        self,
        x: torch.Tensor,
        node_attributes: torch.Tensor,
    ) -> torch.Tensor:
        return self.summation(x, node_attributes)

    # for mypy
    def __call__(
        self,
        x: torch.Tensor,
        node_attributes: torch.Tensor,
    ) -> torch.Tensor:
        return super().__call__(x, node_attributes)


class Contraction(CodeGenMixin, torch.nn.Module):
    def __init__(self, config: ContractionConfig, correlation: int):
        super().__init__()

        self.config = config
        self.correlation = correlation
        self.initial_contraction = InitialContraction(config, correlation)

        self.weight_contractions = UniformModuleList(
            FollowingWeightContraction(config, j)
            for j in reversed(range(1, correlation))
        )

        self.feature_contractions = UniformModuleList(
            FeatureContraction(config, j)
            for j in reversed(range(1, correlation))
        )

    def forward(
        self,
        node_embeddings: torch.Tensor,
        node_attributes: torch.Tensor,
    ) -> torch.Tensor:
        output = self.initial_contraction(node_embeddings, node_attributes)

        for weight_contraction, feature_contraction in zip(
            self.weight_contractions, self.feature_contractions
        ):
            output = weight_contraction(node_attributes) + output
            output = feature_contraction(output, node_embeddings)

        return output.reshape(output.shape[0], -1)

    # for mypy
    def __call__(
        self,
        node_embeddings: torch.Tensor,
        node_attributes: torch.Tensor,
    ) -> torch.Tensor:
        return super().__call__(node_embeddings, node_attributes)

    def __repr__(self) -> str:
        weights = sum(p.numel() for p in self.parameters() if p.requires_grad)
        return (
            f"{self.__class__.__name__}({self.config}, "
            f"correlation={self.correlation}, weights={weights})"
        )


_U_cache_sparse: dict[tuple[str, str, int], torch.Tensor] = torch.load(
    Path(__file__).parent / "_high_order_CG_coeff.pt",
    weights_only=True,
)
"""
A pre-computed look-up table for the U matrices used in MACE Contractions.

Keys are tuples of the form ``(in_irreps, out_irreps, correlation)``, where
``in_irreps`` and ``out_irreps`` are strings formed by concatenating the string
representations of the irreducible representations, and ``correlation`` is an
integer.
"""


def get_U_matrix(
    in_irreps: list[o3.Irrep],
    out_irreps: o3.Irrep,
    correlation: int,
) -> torch.Tensor:
    key = (
        " ".join(map(str, in_irreps)),
        str(out_irreps),
        correlation,
    )
    l_max = correlation * int(o3.Irrep(in_irreps[-1]).l)
    if l_max > 11:
        raise ValueError(
            f"l_max > 11 (you supplied {l_max=}) is not supported by e3nn."
        )
    if key not in _U_cache_sparse:
        raise ValueError(
            f"U_matrix for {key} not found in cache - this is surprising! "
            "Please raise an issue at https://github.com/jla-gardner/graph-pes "
            "so we can fix this."
        )

    return _U_cache_sparse[key].to_dense().to(dtype=torch.get_default_dtype())


class UnflattenIrreps(torch.nn.Module):
    """
    Unflattens a tensor of irreps with uniform multiplicity,
    e.g. (N, 16x0e + 16x1o) -> (N, 16, 1x0e + 1x1o)
    """

    def __init__(self, irreps: list[o3.Irrep], channels: int) -> None:
        super().__init__()
        self.channels = channels
        self.irreps = irreps
        self.dims = [ir.dim for ir in irreps]

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        idx = 0
        output = []

        # iterate over the flat tensor, and pull out
        # each channel x irrep
        # e.g.   (N, 16x0e + 16x1o) -> (N, 16, 1x0e + 1x1o)
        # equiv: (N,      16x(1+3)) -> (N, 16, 1+3)
        for dim in self.dims:
            field = tensor[:, idx : idx + self.channels * dim]
            idx += self.channels * dim
            field = field.reshape(-1, self.channels, dim)
            output.append(field)

        return torch.cat(output, dim=-1)

    def __repr__(self) -> str:
        _in = "(N, "
        _in += "+".join([f"{self.channels}x{ir}" for ir in self.irreps])
        _in += ")"

        _out = f"(N, {self.channels}, "
        _out += "+".join([f"1x{ir}" for ir in self.irreps])
        _out += ")"

        return f"{self.__class__.__name__}({_in} -> {_out})"


def parse_irreps(irreps: str | list[str]) -> list[o3.Irrep]:
    if isinstance(irreps, str):
        try:
            return [ir for _, ir in o3.Irreps(irreps)]
        except ValueError:
            raise ValueError(
                f"Unable to parse {irreps} as irreps. "
                "Expected a string of the form '0e + 1o'"
            ) from None
    try:
        return [o3.Irrep(ir) for ir in irreps]
    except ValueError:
        raise ValueError(
            f"Unable to parse {irreps} as irreps. "
            "Expected a list of strings of the form ['0e', '1o']"
        ) from None



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Any, Literal, Union

import yaml
from pytorch_lightning.loggers import CSVLogger, Logger

from graph_pes.config.shared import TorchConfig
from graph_pes.data import GraphDataset
from graph_pes.training.callbacks import WandbLogger


@dataclass
class TestingConfig:
    """Configuration for testing a GraphPES model."""

    model_path: str
    """The path to the ``model.pt`` file."""

    data: Union[GraphDataset, dict[str, GraphDataset]]  # noqa: UP007
    """
    Either:

    - a single :class:`~graph_pes.data.GraphDataset`. Results will be logged as
      ``"<prefix>/<metric>"``.
    - a mapping from names to datasets. Results will be logged as
      ``"<prefix>/<dataset-name>/<metric>"``, allowing for testing on multiple 
      datasets.
    """

    loader_kwargs: dict[str, Any]
    """
    Keyword arguments to pass to the 
    :class:`~graph_pes.data.loader.GraphDataLoader`.

    Defaults to:

    .. code-block:: yaml

        loader_kwargs:
            batch_size: 2
            num_workers: 0
    
    You should tune this to make testing faster.
    """

    torch: TorchConfig
    """The torch configuration to use for testing."""

    logger: Union[Literal["auto", "csv"], dict[str, Any]] = "auto"  # noqa: UP007
    """
    The logger to use for logging the test metrics.

    If ``"auto"``, we will attempt to find the training config
    from ``<model_path>/../train-config.yaml``, and use the logger
    from that config.

    If ``"csv"``, we will use a CSVLogger.

    If a dictionary, we will instantiate a new 
    :class:`~graph_pes.training.callbacks.WandbLogger` with the provided 
    arguments.
    """

    accelerator: str = "auto"
    """The accelerator to use for testing."""

    prefix: str = "testing"
    """The prefix to use for logging. Individual metrics will be logged as
    ``<prefix>/<dataset_name>/<metric>``.
    """

    def get_logger(self) -> Logger:
        root_dir = Path(self.model_path).parent
        if self.logger == "csv":
            return CSVLogger(save_dir=root_dir, name="")
        elif isinstance(self.logger, dict):
            return WandbLogger(
                output_dir=root_dir,
                log_epoch=False,
                **self.logger,
            )

        if not self.logger == "auto":
            raise ValueError(f"Invalid logger: {self.logger}")

        train_config_path = root_dir / "train-config.yaml"
        if not train_config_path.exists():
            raise ValueError(
                f"Could not find training config at {train_config_path}. "
                "Please specify a logger explicitly."
            )
        with open(train_config_path) as f:
            logger_data = yaml.safe_load(f).get("wandb", None)

        if logger_data is None:
            return CSVLogger(save_dir=root_dir, name="")

        return WandbLogger(output_dir=root_dir, log_epoch=False, **logger_data)

    @classmethod
    def defaults(cls) -> dict:
        return {
            "torch": {"float32_matmul_precision": "high", "dtype": "float32"},
            "loader_kwargs": {"batch_size": 2, "num_workers": 0},
        }



from __future__ import annotations

import logging
import sys
from pathlib import Path

from . import distributed

__all__ = ["logger", "log_to_file", "set_log_level"]


class MultiLineFormatter(logging.Formatter):
    """Detect multi-line logs and adds newlines with colors."""

    def __init__(self):
        super().__init__("[%(name)s %(levelname)s]: %(message)s")

    def format(self, record: logging.LogRecord) -> str:
        record.msg = str(record.msg).strip()
        # add in new lines
        if "\n" in record.msg:
            record.msg = "\n" + record.msg + "\n"
        return super().format(record)


# create the graph-pes logger
logger = logging.getLogger(name="graph-pes")
std_out_handler = logging.StreamHandler(stream=sys.stdout)
std_out_handler.setFormatter(MultiLineFormatter())

# log to stdout if rank 0
if distributed.IS_RANK_0:
    logger.addHandler(std_out_handler)

    # capture all logs but only show INFO and above in stdout (by default)
    logger.setLevel(logging.DEBUG)
    std_out_handler.setLevel(logging.INFO)


def log_to_file(output_dir: str | Path):
    """Append logs to a `rank-<rank>.log` file in the given output directory."""

    file = Path(output_dir) / f"rank-{distributed.GLOBAL_RANK}.log"
    file.parent.mkdir(parents=True, exist_ok=True)

    handler = logging.FileHandler(file, mode="a")
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(MultiLineFormatter())

    logger.addHandler(handler)
    logger.info(f"Logging to {file}")


def set_log_level(level: str | int):
    """Set the logging level."""

    std_out_handler.setLevel(level)
    logger.debug(f"Set logging level to {level}")



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional

import torch
from torch import Tensor

from graph_pes.atomic_graph import (
    DEFAULT_CUTOFF,
    AtomicGraph,
    PropertyKey,
    neighbour_distances,
    sum_over_neighbours,
)
from graph_pes.graph_pes_model import GraphPESModel
from graph_pes.models.components.distances import SmoothOnsetEnvelope
from graph_pes.utils.misc import to_significant_figures, uniform_repr
from graph_pes.utils.nn import PerElementParameter


class PairPotential(GraphPESModel, ABC):
    r"""
    An abstract base class for PES models that calculate total energy as
    a sum over pairwise interactions:

    .. math::
        E = \sum_{i, j} V(r_{ij}, Z_i, Z_j)

    where :math:`r_{ij}` is the distance between atoms :math:`i` and :math:`j`,
    and :math:`Z_i` and :math:`Z_j` are their atomic numbers.
    This can be recast as a sum over local energy contributions,
    :math:`E = \sum_i \varepsilon_i`, according to:

    .. math::
        \varepsilon_i = \frac{1}{2} \sum_j V(r_{ij}, Z_i, Z_j)

    Subclasses should implement :meth:`interaction`.
    """

    def __init__(self, cutoff: float):
        super().__init__(
            cutoff=cutoff,
            implemented_properties=["local_energies"],
        )

    @abstractmethod
    def interaction(
        self, r: torch.Tensor, Z_i: torch.Tensor, Z_j: torch.Tensor
    ) -> torch.Tensor:
        """
        Compute the interactions between pairs of atoms, given their
        distances and atomic numbers.

        Parameters
        ----------
        r
            The pair-wise distances between the atoms.
        Z_i
            The atomic numbers of the central atoms.
        Z_j
            The atomic numbers of the neighbours.

        Returns
        -------
        V: torch.Tensor
            The pair-wise interactions.
        """

    def forward(self, graph: AtomicGraph) -> dict[PropertyKey, Tensor]:
        """
        Predict the local energies as half the sum of the pair-wise
        interactions that each atom participates in.
        """

        # avoid tuple unpacking to keep torchscript happy
        central_atoms = graph.neighbour_list[0]
        neighbours = graph.neighbour_list[1]
        distances = neighbour_distances(graph)

        Z_i = graph.Z[central_atoms]
        Z_j = graph.Z[neighbours]

        V = self.interaction(distances, Z_i, Z_j)  # (E) / (E, 1)

        # sum over the neighbours
        energies = sum_over_neighbours(V.squeeze(), graph)

        # divide by 2 to avoid double counting
        return {"local_energies": energies / 2}


class SmoothedPairPotential(PairPotential):
    r"""
    A wrapper around a :class:`~graph_pes.models.PairPotential` that
    applies a smooth cutoff function, :math:`f` =
    :class:`~graph_pes.models.components.distances.SmoothOnsetEnvelope`,
    to the potential to ensure a continous energy surface:

    .. math::

        V(r) = f(r, r_o, r_c) \cdot V_{\text{wrapped}}(r)

    where

    .. math::

        f(r, r_o, r_c) = \begin{cases}
        \hfill 1 \hfill & \text{if } r < r_o \\
        \frac{(r_c - r)^2 (r_c + 2r - 3r_o)}{(r_c - r_o)^3} & \text{if } r_o \leq r < r_c \\
        \hfill 0 \hfill & \text{if } r \geq r_c
        \end{cases}

    and :math:`r_o` and :math:`r_c` are the onset and cutoff radii respectively.

    Parameters
    ----------
    potential
        The potential to wrap.
    smoothing_onset
        The radius at which the smooth cutoff function begins.
        Defaults to :math:`2r_c / 3`.
    """  # noqa: E501

    def __init__(
        self,
        potential: PairPotential,
        smoothing_onset: float | None = None,
    ):
        cutoff = potential.cutoff.item()
        super().__init__(cutoff=cutoff)
        if not smoothing_onset:
            smoothing_onset = 2 * cutoff / 3
        self.envelope = SmoothOnsetEnvelope(cutoff, smoothing_onset)
        self.potential = potential

    def interaction(self, r: Tensor, Z_i: Tensor, Z_j: Tensor) -> Tensor:
        raw_v = self.potential.interaction(r, Z_i, Z_j)
        return raw_v * self.envelope(r)


class LennardJones(PairPotential):
    r"""
    A pair potential with an interaction term of the form:

    .. math::

        V_{LJ}(r) = 4 \varepsilon \left[ \left( \frac{\sigma}{r} \right)^{12} -
        \left( \frac{\sigma}{r} \right)^{6} \right]

    where :math:`r_{ij}` is the distance between atoms :math:`i` and :math:`j`,
    and :math:`\varepsilon` and :math:`\sigma` are strictly positive
    paramters that control the depth and width of the potential well.

    To ensure a continous energy surface, the final interaction is shifted
    by the value of the potential at the cutoff to remove the discontinuity:

    .. math::

        V(r) = V_{LJ}(r) - V_{LJ}(r_c)

    .. warning::

        The derivative of the potential is not continuous at the cutoff,
        irrespective of whether the potential is shifted or not.
        This leads to discontinuities in this models's forces. Consider wrapping
        this potential in :class:`~graph_pes.models.SmoothedPairPotential`
        to obtain a smooth and continuous potential and force field.

    Parameters
    ----------
    cutoff:
        The cutoff radius.
    epsilon:
        The maximum depth of the potential.
    sigma:
        The distance at which the potential is zero.

    Example
    -------
    .. code-block:: python

        from graph_pes.utils.analysis import dimer_curve
        from graph_pes.models import LennardJones

        dimer_curve(LennardJones(), system="H2", rmax=3.5)

    .. image:: lj-dimer.svg
        :align: center
    """

    def __init__(
        self,
        cutoff: float = DEFAULT_CUTOFF,
        epsilon: float = 0.1,
        sigma: float = 1.0,
        shift: bool = False,
    ):
        super().__init__(cutoff=cutoff)

        self._log_epsilon = torch.nn.Parameter(torch.tensor(epsilon).log())
        self._log_sigma = torch.nn.Parameter(torch.tensor(sigma).log())

        # register buffers for loading from state dict
        self.register_buffer(
            "_offset", self.V_LJ(torch.tensor(cutoff)).detach()
        )
        # save as int to avoid issues with torchscript
        self.register_buffer("_shift", torch.tensor(int(shift)))

    @property
    def epsilon(self) -> torch.Tensor:
        return self._log_epsilon.exp()

    @property
    def sigma(self) -> torch.Tensor:
        return self._log_sigma.exp()

    # don't use Z_i and Z_j, but include them for consistency with the
    # abstract method
    def interaction(
        self,
        r: torch.Tensor,
        Z_i: Optional[torch.Tensor] = None,  # noqa: UP007
        Z_j: Optional[torch.Tensor] = None,  # noqa: UP007
    ):
        v_lj = self.V_LJ(r)
        if self._shift.item():
            return v_lj - self._offset
        return v_lj

    def V_LJ(self, r: torch.Tensor) -> torch.Tensor:
        x = self.sigma / r
        return 4 * self.epsilon * (x**12 - x**6)

    def __repr__(self):
        return uniform_repr(
            self.__class__.__name__,
            epsilon=to_significant_figures(self.epsilon.item(), 3),
            sigma=to_significant_figures(self.sigma.item(), 3),
            cutoff=to_significant_figures(self.cutoff.item(), 3),
        )

    @staticmethod
    def from_ase(
        sigma: float = 1.0,
        epsilon: float = 1.0,
        rc: float | None = None,
        ro: float | None = None,
        smooth: bool = False,
    ):
        """
        Create a :class:`LennardJones` potential with an interface
        identical to the ASE :class:`ase.calculators.lj.LennardJones`
        calculator.

        Please refer to the ASE documentation for more details.

        Parameters
        ----------
        sigma
            The distance at which the potential is zero.
        epsilon
            The maximum depth of the potential.
        rc
            The cutoff radius. If not given, the default value is 3 * sigma.
        ro
            The radius at which the smooth cutoff function begins.
        """
        if rc is None:
            rc = 3 * sigma
        if ro is None:
            ro = 2 * rc / 3

        if not smooth:
            return LennardJones(rc, epsilon, sigma, shift=True)
        return SmoothedPairPotential(
            LennardJones(rc, epsilon, sigma, shift=False), ro
        )


class Morse(PairPotential):
    r"""
    A pair potential of the form:

    .. math::
        V(r_{ij}, Z_i, Z_j) = V(r_{ij}) = D (1 - e^{-a(r_{ij} - r_0)})^2

    where :math:`r_{ij}` is the distance between atoms :math:`i` and :math:`j`,
    and :math:`D`, :math:`a` and :math:`r_0` are strictly positive parameters
    that control the depth, width and center of the potential well respectively.

    Parameters
    ----------
    D:
        The maximum depth of the potential.
    a:
        A measure of the width of the potential.
    r0:
        The distance at which the potential is at its minimum.

    Example
    -------
    .. code-block:: python

        from graph_pes.utils.analysis import dimer_curve
        from graph_pes.models import Morse

        dimer_curve(Morse(), system="H2", rmax=3.5)

    .. image:: morse-dimer.svg
        :align: center
    """

    def __init__(
        self,
        cutoff: float = DEFAULT_CUTOFF,
        D: float = 0.1,
        a: float = 5.0,
        r0: float = 1.5,
    ):
        super().__init__(cutoff=cutoff)

        self._log_D = torch.nn.Parameter(torch.tensor(D).log())
        self._log_a = torch.nn.Parameter(torch.tensor(a).log())
        self._log_r0 = torch.nn.Parameter(torch.tensor(r0).log())

    @property
    def D(self):
        return self._log_D.exp()

    @property
    def a(self):
        return self._log_a.exp()

    @property
    def r0(self):
        return self._log_r0.exp()

    def interaction(
        self,
        r: torch.Tensor,
        Z_i: Optional[torch.Tensor] = None,  # noqa: UP007
        Z_j: Optional[torch.Tensor] = None,  # noqa: UP007
    ):
        """
        Evaluate the pair potential.

        Parameters
        ----------
        r : torch.Tensor
            The pair-wise distances between the atoms.
        Z_i : torch.Tensor
            The atomic numbers of the central atoms. (unused)
        Z_j : torch.Tensor
            The atomic numbers of the neighbours. (unused)
        """
        return self.D * (1 - torch.exp(-self.a * (r - self.r0))) ** 2

    def pre_fit(self, graphs: AtomicGraph):
        # set the center of the well to be close to the minimum pair-wise
        # distance: the 10th percentile plus a small offset
        d = torch.quantile(neighbour_distances(graphs), 0.1) + 0.1
        self._log_r0 = torch.nn.Parameter(d.log())

    def __repr__(self):
        return uniform_repr(
            self.__class__.__name__,
            D=to_significant_figures(self.D.item(), 3),
            a=to_significant_figures(self.a.item(), 3),
            r0=to_significant_figures(self.r0.item(), 3),
        )


class LennardJonesMixture(PairPotential):
    r"""
    An extension of the simple :class:`LennardJones` potential to
    account for multiple atomic species.

    Each element is associated with a unique pair of parameters, $\sigma_i$ and
    $\varepsilon_i$, which control the width and depth of the potential well
    for that element.

    Interactions between atoms of different elements are calculated using
    effective parameters :math:`\sigma_{i\neq j}` and :math:`\varepsilon_{i
    \neq j}`, which are calculated as:

    * :math:`\sigma_{i\neq j} = \nu_{ij} \cdot (\sigma_i + \sigma_j) / 2`
    * :math:`\varepsilon_{i\neq j} = \zeta_{ij} \cdot \sqrt{\varepsilon_i
      \cdot \varepsilon_j}`

    where :math:`\nu_{ij}` is a mixing parameter that controls the width of the
    potential well.

    For more details, see `wikipedia <https://en.wikipedia.org/wiki/Lennard-Jones_potential#Mixtures_of_Lennard-Jones_substances>`_.
    """

    def __init__(
        self,
        cutoff: float = DEFAULT_CUTOFF,
        modulate_distances: bool = True,
    ):
        super().__init__(cutoff=cutoff)

        self.modulate_distances: Tensor
        self.register_buffer(
            "modulate_distances", torch.tensor(int(modulate_distances))
        )

        self.epsilon = PerElementParameter.of_length(1, default_value=0.1)
        self.sigma = PerElementParameter.covalent_radii(scaling_factor=0.9)
        self.nu = PerElementParameter.of_length(
            1, index_dims=2, default_value=1
        )
        self.zeta = PerElementParameter.of_length(
            1, index_dims=2, default_value=1
        )

    def interaction(self, r: Tensor, Z_i: Tensor, Z_j: Tensor) -> Tensor:
        """
        Evaluate the pair potential.

        Parameters
        ----------
        r : torch.Tensor
            The pair-wise distances between the atoms.
        Z_i : torch.Tensor
            The atomic numbers of the central atoms.
        Z_j : torch.Tensor
            The atomic numbers of the neighbours.
        """
        cross_interaction = Z_i != Z_j  # (E)

        sigma_j = self.sigma[Z_j].squeeze()  # (E)
        sigma_i = self.sigma[Z_i].squeeze()  # (E)
        nu = (
            self.nu[Z_i, Z_j].squeeze()
            if self.modulate_distances
            else torch.tensor(1)
        )
        sigma = torch.where(
            cross_interaction,
            nu * (sigma_i + sigma_j) / 2,
            sigma_i,
        ).clamp(min=0.2)  # (E)

        epsilon_i = self.epsilon[Z_i].squeeze()
        epsilon_j = self.epsilon[Z_j].squeeze()
        zeta = self.zeta[Z_i, Z_j].squeeze()
        epsilon = torch.where(
            cross_interaction,
            zeta * (epsilon_i * epsilon_j).sqrt(),
            epsilon_i,
        ).clamp(min=0.00)  # (E)

        x = sigma / r
        return 4 * epsilon * (x**12 - x**6)

    def __repr__(self):
        kwargs = {
            "sigma": self.sigma,
            "epsilon": self.epsilon,
            "zeta": self.zeta,
        }
        if self.modulate_distances:
            kwargs["nu"] = self.nu

        return uniform_repr(
            self.__class__.__name__,
            **kwargs,
            indent_width=2,
            max_width=60,
            stringify=False,
        )


class ZBLCoreRepulsion(PairPotential):
    r"""
    The `ZBL <https://en.wikipedia.org/wiki/Stopping_power_(particle_radiation)#Repulsive_interatomic_potentials>`__ repulsive potential:

    .. math::

        V(r, Z_i, Z_j) = \frac{e^2}{4 \pi \epsilon_0} \cdot \frac{Z_i Z_j}{r} \cdot \phi(r / a)

    where :math:`\phi(x)` is a dimensionless function given by:

    .. math::

        \phi(x) = 0.1818e^{-3.2x} + 0.5099e^{-0.9423x} + 0.2802e^{-0.4029x} + 0.02817e^{-0.2016x}

    and :math:`a` is the screening length:

    .. math::

        a = \frac{\lambda_p \cdot a_0}{Z_i^{\lambda_e} + Z_j^{\lambda_e}}

    where :math:`a_0` is the Bohr radius, :math:`\lambda_p = 0.8854` and
    :math:`\lambda_e = 0.23`.

    Parameters
    ----------
    cutoff : float, optional
        The cutoff radius for the potential. Default is DEFAULT_CUTOFF.
    trainable : bool, optional
        If True, the pre-factor (:math:`\lambda_p`) and exponent
        (:math:`\lambda_e`) of the screening length are trainable parameters.
        Default is False.

    Example
    -------
    .. code-block:: python

        import matplotlib.pyplot as plt
        from graph_pes.utils.analysis import dimer_curve
        from graph_pes.models import ZBL

        dimer_curve(ZBL(), system="H2", rmin=0.1, rmax=3.5)
        plt.xlim(0, 3.5)
        plt.ylim(0.01, 100)
        plt.yscale("log")

    .. image:: zbl-dimer.svg
        :align: center
    """  # noqa: E501

    def __init__(self, cutoff: float = DEFAULT_CUTOFF, trainable: bool = False):
        super().__init__(cutoff=cutoff)

        if trainable:
            self.pre_factor = torch.nn.Parameter(torch.tensor(0.8854))
            self.exponent = torch.nn.Parameter(torch.tensor(0.23))
        else:
            self.register_buffer("pre_factor", torch.tensor(0.8854))
            self.register_buffer("exponent", torch.tensor(0.23))

        ZBL_CONSTANTS = {
            "coeff": [0.1818, 0.5099, 0.2802, 0.02817],
            "exp": [-3.2, -0.9423, -0.4029, -0.2016],
        }
        self.register_buffer("ZBL_coeff", torch.tensor(ZBL_CONSTANTS["coeff"]))
        self.register_buffer("ZBL_exp", torch.tensor(ZBL_CONSTANTS["exp"]))

    def interaction(self, r: Tensor, Z_i: Tensor, Z_j: Tensor) -> Tensor:
        BOHR_RADIUS = 0.529177249  # Å
        COULOMB_CONSTANT = 14.3996  # eV Å

        a = (
            self.pre_factor
            * BOHR_RADIUS
            / (torch.pow(Z_i, self.exponent) + torch.pow(Z_j, self.exponent))
        )
        x = r / a
        phi = (
            self.ZBL_coeff[0] * torch.exp(self.ZBL_exp[0] * x)
            + self.ZBL_coeff[1] * torch.exp(self.ZBL_exp[1] * x)
            + self.ZBL_coeff[2] * torch.exp(self.ZBL_exp[2] * x)
            + self.ZBL_coeff[3] * torch.exp(self.ZBL_exp[3] * x)
        )

        return COULOMB_CONSTANT * Z_i * Z_j * phi / r



from __future__ import annotations

from functools import reduce
from typing import (
    Any,
    Generic,
    Iterable,
    Iterator,
    Sequence,
    TypedDict,
    TypeVar,
)

import torch
import torch.nn
from ase.data import atomic_numbers, chemical_symbols, covalent_radii
from torch import Tensor

from graph_pes.utils.misc import left_aligned_mul

from .misc import (
    MAX_Z,
    pairs,
    to_significant_figures,
    uniform_repr,
)

V = TypeVar("V", bound=torch.nn.Module)


class UniformModuleDict(torch.nn.ModuleDict, Generic[V]):
    """
    A :class:`torch.nn.ModuleDict` sub-class for cases where
    the values are all of the same type.

    Examples
    --------
    >>> from graph_pes.utils.nn import UniformModuleDict
    >>> from torch.nn import Linear
    >>> linear_dict = UniformModuleDict(a=Linear(10, 5), b=Linear(5, 1))
    """

    def __init__(self, **modules: V):
        super().__init__(modules)

    def values(self) -> Iterable[V]:
        return super().values()  # type: ignore

    def items(self) -> Iterable[tuple[str, V]]:
        return super().items()  # type: ignore

    def __getitem__(self, key: str) -> V:
        return super().__getitem__(key)  # type: ignore

    def __setitem__(self, key: str, module: V) -> None:  # type: ignore
        super().__setitem__(key, module)

    def pop(self, key: str) -> V:
        return super().pop(key)  # type: ignore


class UniformModuleList(torch.nn.ModuleList, Sequence[V]):
    """
    A :class:`torch.nn.ModuleList` sub-class for cases where
    the values are all of the same type.

    Examples
    --------
    >>> from graph_pes.utils.nn import UniformModuleList
    >>> from torch.nn import Linear
    >>> linear_list = UniformModuleList(Linear(10, 5), Linear(5, 1))
    """

    def __init__(self, modules: Iterable[V]):
        super().__init__(modules)

    def __getitem__(self, idx: int) -> V:  # type: ignore
        return super().__getitem__(idx)  # type: ignore

    def __setitem__(self, idx: int, value: V) -> None:  # type: ignore
        super().__setitem__(idx, value)

    def append(self, module: V) -> None:  # type: ignore
        super().append(module)

    def extend(self, modules: Iterable[V]) -> None:  # type: ignore
        super().extend(modules)

    def insert(self, idx: int, module: V) -> None:  # type: ignore
        super().insert(idx, module)

    def pop(self, idx: int) -> V:  # type: ignore
        return super().pop(idx)  # type: ignore

    def __iter__(self) -> Iterator[V]:
        return super().__iter__()  # type: ignore


class MLPConfig(TypedDict):
    """
    A TypedDict helper class for configuring an :class:`MLP`.

    Examples
    --------
    Specify this in a config file:

    .. code-block:: yaml

        mlp:
            hidden_depth: 3
            hidden_features: 64
            activation: SiLU
    """

    hidden_depth: int
    """The number of hidden layers in the MLP."""
    hidden_features: int
    """The number of features in the hidden layers."""
    activation: str
    """The activation function to use."""


class MLP(torch.nn.Module):
    """
    A multi-layer perceptron model, alternating linear layers and activations.

    Parameters
    ----------
    layers
        The number of nodes in each layer.
    activation
        The activation function to use: either a named activation function
        from `torch.nn`, or a `torch.nn.Module` instance.
    activate_last
        Whether to apply the activation function after the last linear layer.
    bias
        Whether to include bias terms in the linear layers.

    Examples
    --------
    >>> import torch
    >>> from graph_pes.utils.nn import MLP
    >>> model = MLP([10, 5, 1])
    >>> model
    MLP(10 → 5 → 1, activation=CELU())
    >>> MLP([10, 5, 1], activation=torch.nn.ReLU())
    MLP(10 → 5 → 1, activation=ReLU())
    >>> MLP([10, 5, 1], activation="Tanh")
    MLP(10 → 5 → 1, activation=Tanh())
    """

    def __init__(
        self,
        layers: list[int],
        activation: str | torch.nn.Module = "CELU",
        activate_last: bool = False,
        bias: bool = True,
    ):
        super().__init__()

        self.activation = (
            parse_activation(activation)
            if isinstance(activation, str)
            else activation
        )
        self.activate_last = activate_last

        self.linear_layers = torch.nn.ModuleList(
            [
                torch.nn.Linear(_in, _out, bias=bias)
                for _in, _out in pairs(layers)
            ]
        )

    def forward(self, x: Tensor) -> Tensor:
        """
        Perform a forward pass through the network.

        Parameters
        ----------
        x
            The input to the network.
        """
        for i, linear in enumerate(self.linear_layers):
            x = linear(x)
            last_layer = i == len(self.linear_layers) - 1
            if not last_layer or self.activate_last:
                x = self.activation(x)

        return x

    @property
    def input_size(self):
        """The size of the input to the network."""
        return self.linear_layers[0].in_features

    @property
    def output_size(self):
        """The size of the output of the network."""
        return self.linear_layers[-1].out_features

    @property
    def layer_widths(self):
        """The widths of the layers in the network."""
        inputs = [layer.in_features for layer in self.linear_layers]
        return inputs + [self.output_size]

    def __repr__(self):
        layers = " → ".join(map(str, self.layer_widths))
        return uniform_repr(
            self.__class__.__name__,
            layers,
            activation=self.activation,
            stringify=False,
        )

    @classmethod
    def from_config(
        cls,
        config: MLPConfig,
        input_features: int,
        output_features: int,
        bias: bool = True,
    ) -> MLP:
        """
        Create an :class:`MLP` from a configuration.
        """
        return cls(
            layers=[input_features]
            + [config["hidden_features"]] * config["hidden_depth"]
            + [output_features],
            activation=config["activation"],
            bias=bias,
        )


class ShiftedSoftplus(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.shift = torch.log(torch.tensor(2.0)).item()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.nn.functional.softplus(x) - self.shift

    def __repr__(self):
        return uniform_repr(self.__class__.__name__)


def parse_activation(act: str) -> torch.nn.Module:
    """
    Parse a string into a PyTorch activation function.

    Parameters
    ----------
    act
        The activation function to parse.

    Returns
    -------
    torch.nn.Module
        The parsed activation function.
    """
    activation = getattr(torch.nn, act, None)
    if activation is None:
        raise ValueError(f"Activation function {act} not found in `torch.nn`.")
    return activation()


def prod(iterable):
    return reduce(lambda x, y: x * y, iterable, 1)


class PerElementParameter(torch.nn.Parameter):
    """
    A subclass of :class:`torch.nn.Parameter` that is indexed by atomic
    number/s. Crucially, this subclass overrides the :meth:`numel` method,
    for accurately counting the number of relevant and learnable parameters.

    Examples
    --------
    Imagine the case where you have a model parameter with a value for each
    element in the periodic table. If you only train the model on a dataset
    containing a few elements, you don't want to count the total number of
    parameters, as this will be unnecessarily large.

    >>> # don't do this!
    >>> per_element_parameter = torch.nn.Parameter(torch.randn(119))
    >>> per_element_parameter.numel()
    119
    >>> per_element_parameter
    Parameter containing:
    tensor([ 1.2838e-01, -1.4507e+00,  1.3465e-01, -9.5786e-01, ...,
            -1.3329e+00, -1.5515e+00,  2.1106e+00, -9.7268e-01],
       requires_grad=True)

    >>> # do this instead
    >>> per_element_paramter = PerElementParameter.of_shape((1,))
    >>> per_element_parameter.register_elements([1, 6, 8])
    >>> per_element_parameter.numel()
    3
    >>> per_element_parameter
    PerElementParameter({'O': -0.278, 'H': 0.157, 'C': -0.0379}, trainable=True)

    ``graph-pes-train`` automatically registers all elements that a model
    encounters during training, so you rarely need to call
    :meth:`register_elements` yourself.
    """

    def __new__(
        cls, data: Tensor, requires_grad: bool = True
    ) -> PerElementParameter:
        pep = super().__new__(cls, data, requires_grad=requires_grad)
        pep._is_per_element_param = True  # type: ignore
        return pep  # type: ignore

    def __init__(self, data: Tensor, requires_grad: bool = True):
        super().__init__()
        # set extra state
        self._accessed_Zs: set[int] = set()
        # set this to an arbitrary value: this gets updated post-init
        self._index_dims: int = 1

    def register_elements(self, Zs: Iterable[int]) -> None:
        """
        Register the elements that are relevant for the parameter.

        This is typically only used internally - you shouldn't call this
        yourself in any of your model definitions.
        """
        self._accessed_Zs.update(sorted(Zs))

    @classmethod
    def of_shape(
        cls,
        shape: tuple[int, ...] = (),
        index_dims: int = 1,
        default_value: float | None = None,
        requires_grad: bool = True,
    ) -> PerElementParameter:
        """
        Create a :class:`PerElementParameter` with a given shape for each
        element in the periodic table.

        Parameters
        ----------
        shape
            The shape of the parameter for each element.
        index_dims
            The number of dimensions to index by.
        default_value
            The value to initialise the parameter with. If ``None``, the
            parameter is initialised with random values.
        requires_grad
            Whether the parameter should be learnable.

        Returns
        -------
        PerElementParameter
            The created parameter.

        Examples
        --------
        Create a parameter intended to be indexed by a single atomic number,
        i.e. ``pep[Z]``:

        >>> PerElementParameter.of_shape((3,)).shape
        torch.Size([119, 3])
        >>> PerElementParameter.of_shape((3, 4)).shape
        torch.Size([119, 3, 4])

        Create a parameter intended to be indexed by two atomic numbers, i.e.
        ``pep[Z1, Z2]``:

        >>> PerElementParameter.of_shape((3,), index_dims=2).shape
        torch.Size([119, 119, 3])
        """
        actual_shape = tuple([MAX_Z + 1] * index_dims) + shape
        if default_value is not None:
            data = torch.full(actual_shape, float(default_value))
        else:
            data = torch.randn(actual_shape)
        psp = PerElementParameter(data, requires_grad=requires_grad)
        psp._index_dims = index_dims
        return psp

    @classmethod
    @torch.no_grad()
    def from_dict(
        cls,
        requires_grad: bool = True,
        default_value: float = 0.0,
        **values: float,
    ) -> PerElementParameter:
        """
        Create a :class:`PerElementParameter` containing a single value for
        each element in the periodic table from a dictionary of values.

        Parameters
        ----------
        requires_grad
            Whether the parameter should be learnable.
        default_value
            The value to initialise the parameter with. If ``None``, the
            parameter is initialised with random values.
        values
            A dictionary of values, indexed by element symbol.

        Returns
        -------
        PerElementParameter
            The created parameter.

        Examples
        --------
        >>> from graph_pes.utils.nn import PerElementParameter
        >>> pep = PerElementParameter.from_dict(H=1.0, O=2.0)
        >>> pep.register_elements([1, 6, 8])
        >>> pep
        PerElementParameter({'H': 1.0, 'C': 0.0, 'O': 2.0}, trainable=True)
        """
        pep = PerElementParameter.of_length(
            1, requires_grad=requires_grad, default_value=default_value
        )
        for element_symbol, value in values.items():
            if element_symbol not in chemical_symbols:
                raise ValueError(f"Unknown element: {element_symbol}")
            Z = chemical_symbols.index(element_symbol)
            pep[Z] = value

        pep.register_elements(atomic_numbers[v] for v in values)

        return pep

    @classmethod
    def of_length(
        cls,
        length: int,
        index_dims: int = 1,
        default_value: float | None = None,
        requires_grad: bool = True,
    ) -> PerElementParameter:
        """
        Alias for ``PerElementParameter.of_shape((length,), **kwargs)``.
        """
        return PerElementParameter.of_shape(
            (length,), index_dims, default_value, requires_grad
        )

    @classmethod
    @torch.no_grad()
    def covalent_radii(
        cls,
        scaling_factor: float = 1.0,
    ) -> PerElementParameter:
        """
        Create a :class:`PerElementParameter` containing the covalent radii of
        each element in the periodic table.
        """
        pep = PerElementParameter.of_length(1, default_value=1.0)
        for Z in range(1, MAX_Z + 1):
            pep[Z] = torch.tensor(covalent_radii[Z]) * scaling_factor
        return pep

    def numel(self) -> int:
        n_elements = len(self._accessed_Zs)
        accessed_parameters = n_elements**self._index_dims
        per_element_size = prod(self.shape[self._index_dims :])
        return accessed_parameters * per_element_size

    # needed for de/serialization
    def __reduce_ex__(self, proto):
        return (
            _rebuild_per_element_parameter,
            (self.data, self.requires_grad, torch._utils._get_obj_state(self)),
        )

    def __instancecheck__(self, instance) -> bool:
        return super().__instancecheck__(instance) or (  # type: ignore[no-untyped-call]
            isinstance(instance, torch.Tensor)
            and getattr(instance, "_is_per_element_param", False)
        )

    @torch.no_grad()
    def _repr(
        self,
        alias: str | None = None,
        more_info: dict[str, Any] | None = None,
    ) -> str:
        alias = alias or self.__class__.__name__
        more_info = more_info or {}
        if "trainable" not in more_info:
            more_info["trainable"] = self.requires_grad

        if len(self._accessed_Zs) == 0:
            if self._index_dims == 1 and self.shape[1] == 1:
                return uniform_repr(alias, **more_info)

            return uniform_repr(
                alias,
                index_dims=self._index_dims,
                shape=tuple(self.shape[self._index_dims :]),
                **more_info,
            )

        if self._index_dims == 1:
            if self.shape[1] == 1:
                d = {
                    chemical_symbols[Z]: to_significant_figures(self[Z].item())
                    for Z in self._accessed_Zs
                }
                string = f"{alias}({str(d)}, "
                for k, v in more_info.items():
                    string += f"{k}={v}, "
                return string[:-2] + ")"

            elif len(self.shape) == 2:
                d = {
                    chemical_symbols[Z]: self[Z].tolist()
                    for Z in self._accessed_Zs
                }
                string = f"{alias}({str(d)}, "
                for k, v in more_info.items():
                    string += f"{k}={v}, "
                return string[:-2] + ")"

        if self._index_dims == 2 and self.shape[2] == 1:
            columns = []
            columns.append(
                ["Z"] + [chemical_symbols[Z] for Z in self._accessed_Zs]
            )
            for col_Z in self._accessed_Zs:
                row: list[str | float] = [chemical_symbols[col_Z]]
                for row_Z in self._accessed_Zs:
                    row.append(
                        to_significant_figures(self[col_Z, row_Z].item())
                    )
                columns.append(row)

            widths = [max(len(str(x)) for x in col) for col in zip(*columns)]
            lines = []
            for row in columns:
                line = ""
                for x, w in zip(row, widths):
                    # right align
                    line += f"{x:>{w}}  "
                lines.append(line)
            table = "\n" + "\n".join(lines)
            return uniform_repr(
                alias,
                table,
                **more_info,
            )

        return uniform_repr(
            alias,
            index_dims=self._index_dims,
            accessed_Zs=sorted(self._accessed_Zs),
            shape=tuple(self.shape[self._index_dims :]),
            **more_info,
        )

    def __repr__(self, *, tensor_contents=None):
        return self._repr()


def _rebuild_per_element_parameter(data, requires_grad, state):
    psp = PerElementParameter(data, requires_grad)
    psp._accessed_Zs = state["_accessed_Zs"]
    psp._index_dims = state["_index_dims"]
    return psp


class PerElementEmbedding(torch.nn.Module):
    """
    A per-element equivalent of :class:`torch.nn.Embedding`.

    Parameters
    ----------
    dim
        The length of each embedding vector.

    Examples
    --------
    >>> embedding = PerElementEmbedding(10)
    >>> len(graph["atomic_numbers"])  # number of atoms in the graph
    24
    >>> embedding(graph["atomic_numbers"])
    <tensor of shape (24, 10)>
    """

    def __init__(self, dim: int):
        super().__init__()
        self._embeddings = PerElementParameter.of_length(dim)

    def forward(self, Z: Tensor) -> Tensor:
        return self._embeddings[Z]

    def dim(self) -> int:
        return self._embeddings.shape[1]

    def __repr__(self) -> str:
        Zs = sorted(self._embeddings._accessed_Zs)
        return uniform_repr(
            self.__class__.__name__,
            dim=self._embeddings.shape[1],
            elements=[chemical_symbols[Z] for Z in Zs],
        )

    def __call__(self, Z: Tensor) -> Tensor:
        return super().__call__(Z)


class HaddamardProduct(torch.nn.Module):
    def __init__(
        self, *components: torch.nn.Module, left_aligned: bool = False
    ):
        super().__init__()
        self.components: list[torch.nn.Module] = torch.nn.ModuleList(components)  # type: ignore
        self.left_aligned = left_aligned

    def forward(self, x):
        out = torch.scalar_tensor(1)
        for component in self.components:
            if self.left_aligned:
                out = left_aligned_mul(out, component(x))
            else:
                out = out * component(x)
        return out


def learnable_parameters(module: torch.nn.Module) -> int:
    """Count the number of **learnable** parameters a module has."""
    return sum(p.numel() for p in module.parameters() if p.requires_grad)


class AtomicOneHot(torch.nn.Module):
    """
    Takes a tensor of atomic numbers Z, and returns a one-hot encoding of
    the atomic numbers.

    Parameters
    ----------
    n_elements
        The total number of expected atomic numbers.
    """

    def __init__(self, elements: list[str]):
        super().__init__()

        self.elements = elements
        self.n_elements = len(elements)

        self.Z_to_idx: Tensor
        self.register_buffer(
            "Z_to_idx",
            # deliberately crazy value to catch errors
            torch.full((MAX_Z + 1,), fill_value=1234),
        )
        for i, symbol in enumerate(elements):
            Z = atomic_numbers[symbol]
            self.Z_to_idx[Z] = i

        # make this a local variable to keep torchscript happy
        self.chemical_symbols: list[str] = chemical_symbols

    def forward(self, Z: Tensor) -> Tensor:
        internal_idx = self.Z_to_idx[Z]

        with torch.no_grad():
            if (internal_idx == 1234).any():
                unknown_Z = torch.unique(Z[internal_idx == 1234])
                unknown_symbols = [self.chemical_symbols[Z] for Z in unknown_Z]

                raise ValueError(
                    f"Unknown elements: {unknown_symbols}. "
                    f"Expected one of {self.elements}"
                )

        return torch.nn.functional.one_hot(
            internal_idx, self.n_elements
        ).float()

    def __repr__(self):
        return uniform_repr(
            self.__class__.__name__,
            elements=self.elements,
        )



from __future__ import annotations

import os
from datetime import datetime
from pathlib import Path

import pytorch_lightning as pl
import yaml
from pytorch_lightning.callbacks import (
    LearningRateMonitor,
    ModelCheckpoint,
    ProgressBar,
    RichProgressBar,
)
from pytorch_lightning.loggers import CSVLogger

from graph_pes.config.shared import (
    instantiate_config_from_dict,
    parse_loss,
    parse_model,
)
from graph_pes.config.training import EarlyStoppingConfig, TrainingConfig
from graph_pes.scripts.utils import (
    configure_general_options,
    extract_config_dict_from_command_line,
    update_summary,
)
from graph_pes.training.callbacks import (
    EarlyStoppingWithLogging,
    GraphPESCallback,
    LoggedProgressBar,
    ModelTimer,
    OffsetLogger,
    SaveBestModel,
    ScalesLogger,
    WandbLogger,
)
from graph_pes.training.tasks import test_with_lightning, train_with_lightning
from graph_pes.training.utils import VALIDATION_LOSS_KEY
from graph_pes.utils import distributed
from graph_pes.utils.logger import log_to_file, logger, set_log_level
from graph_pes.utils.misc import random_dir


def train_from_config(config_data: dict):
    """
    Train a model from a nested dictionary of config.

    We let PyTorch Lightning automatically detect and spin up the
    distributed training run if available.

    Parameters
    ----------
    config
        The configuration data.
    """
    # If we are running in a single process setting, we are always rank 0.
    # This script will run from top to bottom as normal.
    #
    # We allow for automated handling of distributed training runs. Since
    # PyTorch Lightning can do this for us automatically, we delegate all
    # DDP setup to them.
    #
    # In the distributed setting, the order of events is as follows:
    # 1. the user runs `graph-pes-train`, launching a single process on the
    #    global rank 0 process.
    # 2. (on rank 0) this script runs through until the `trainer.fit` is called
    # 3. (on rank 0) the trainer spins up the DDP backend on this process and
    #    launches the remaining processes
    # 4. (on all non-0 ranks) this script runs again until `trainer.fit` is hit.
    # 5. (on all ranks) the trainer sets up the distributed backend and
    #    synchronizes the GPUs: training then proceeds as normal.

    _log_level = config_data.get("general", {}).get("log_level", "INFO")
    set_log_level(_log_level)

    now_ms = datetime.now().strftime("%F %T.%f")[:-3]
    logger.info(f"Started `graph-pes-train` at {now_ms}")

    logger.debug("Parsing config...")

    config_data, config = instantiate_config_from_dict(
        config_data, TrainingConfig
    )
    logger.info("Successfully parsed config.")

    # generate / look up the output directory for this training run
    # and handle the case where there is an ID collision by incrementing
    # the version number
    if distributed.IS_RANK_0:
        # set up directory structure
        if config.general.run_id is None:
            output_dir = random_dir(Path(config.general.root_dir))
        else:
            output_dir = Path(config.general.root_dir) / config.general.run_id
            version = 0
            while output_dir.exists():
                version += 1
                output_dir = (
                    Path(config.general.root_dir)
                    / f"{config.general.run_id}-{version}"
                )

            if version > 0:
                logger.warning(
                    f'Specified run ID "{config.general.run_id}" already '
                    f"exists. Using {output_dir.name} instead."
                )

        output_dir.mkdir(parents=True)
        with open(output_dir / "train-config.yaml", "w") as f:
            yaml.dump(config_data, f)

        # communicate the output directory to other ranks
        distributed.send_to_other_ranks("OUTPUT_DIR", str(output_dir))

    else:
        # get the output directory from rank 0
        output_dir = Path(distributed.receive_from_rank_0("OUTPUT_DIR"))

    log_to_file(output_dir)
    # use the updated output_dir to extract the actual run ID:
    config_data["general"]["run_id"] = output_dir.name
    config.general.run_id = output_dir.name
    logger.info(f"ID for this training run: {config.general.run_id}")
    logger.info(f"""\
Output for this training run can be found at:
   └─ {output_dir}
      ├─ rank-0.log         # find a verbose log here
      ├─ model.pt           # the best model (according to valid/loss/total)
      ├─ lammps_model.pt    # the best model deployed to LAMMPS
      ├─ train-config.yaml  # the complete config used for this run
      └─ summary.yaml       # the summary of the training run
""")

    configure_general_options(config.general.torch, config.general.seed)

    trainer = trainer_from_config(config, output_dir)

    assert trainer.logger is not None
    trainer.logger.log_hyperparams(config_data)

    # instantiate and log things
    model = parse_model(config.model)
    logger.debug(f"Model:\n{model}")

    data = config.get_data()
    logger.debug(f"Data:\n{data}")

    optimizer = config.fitting.optimizer
    logger.debug(f"Optimizer:\n{optimizer}")

    scheduler = config.fitting.scheduler
    _scheduler_str = scheduler if scheduler is not None else "No LR scheduler."
    logger.debug(f"Scheduler:\n{_scheduler_str}")

    total_loss = parse_loss(config.loss)
    logger.debug(f"Total loss:\n{total_loss}")

    train_with_lightning(
        trainer,
        model,
        data,
        loss=total_loss,
        fit_config=config.fitting,
        optimizer=optimizer,
        scheduler=scheduler,
        user_eval_metrics=[],
    )
    update_summary(trainer.logger, output_dir / "summary.yaml")
    logger.info("Training complete.")

    logger.info("Testing best model...")
    test_datasets = {
        "train": data.train,
        "valid": data.valid,
    }
    if config.data.test is not None:
        if isinstance(config.data.test, dict):
            test_datasets.update(config.data.test)
        else:
            test_datasets["test"] = config.data.test

    if isinstance(trainer.logger, WandbLogger):
        trainer.logger._log_epoch = False

    tester = pl.Trainer(
        logger=trainer.logger,
        accelerator=trainer.accelerator,
        inference_mode=False,
    )
    test_with_lightning(
        tester,
        model,
        test_datasets,
        loader_kwargs=config.fitting.loader_kwargs,
        logging_prefix="best_model",
        user_eval_metrics=[],
    )
    update_summary(tester.logger, output_dir / "summary.yaml")

    logger.info("Testing complete.")
    logger.info("Awaiting final Lightning and W&B shutdown...")


def trainer_from_config(
    config: TrainingConfig,
    output_dir: Path,
) -> pl.Trainer:
    # set up a logger on every rank - PTL handles this gracefully so that
    # e.g. we don't spin up >1 wandb experiment
    if config.wandb is not None:
        lightning_logger = WandbLogger(
            output_dir, log_epoch=True, **config.wandb
        )
    else:
        lightning_logger = CSVLogger(save_dir=output_dir, name="")
    logger.debug(f"Logging using {lightning_logger}")

    # create the trainer
    trainer_kwargs = {**config.fitting.trainer_kwargs}
    trainer_kwargs["inference_mode"] = False

    # set up the callbacks
    callbacks = trainer_kwargs.pop("callbacks", [])
    callbacks.extend(config.fitting.callbacks)

    for klass in [OffsetLogger, ScalesLogger, ModelTimer, SaveBestModel]:
        if not any(isinstance(cb, klass) for cb in callbacks):
            callbacks.append(klass())
    if config.fitting.swa is not None:
        callbacks.append(config.fitting.swa.instantiate_lightning_callback())
    if not any(isinstance(c, ProgressBar) for c in callbacks):
        callbacks.append(
            {"rich": RichProgressBar(), "logged": LoggedProgressBar()}[
                config.general.progress
            ]
        )
    if not any(isinstance(c, LearningRateMonitor) for c in callbacks):
        callbacks.append(LearningRateMonitor(logging_interval="epoch"))

    early_stopping_config = None
    # TODO: remove this at a later date
    if config.fitting.early_stopping_patience is not None:
        logger.warning(
            "`early_stopping_patience` is deprecated. Use the `early_stopping` "
            "config option instead.",
        )
        early_stopping_config = EarlyStoppingConfig(
            patience=config.fitting.early_stopping_patience,
            min_delta=1e-6,
            monitor=VALIDATION_LOSS_KEY,
        )
    if config.fitting.early_stopping is not None:
        early_stopping_config = config.fitting.early_stopping

    if early_stopping_config is not None:
        callbacks.append(
            EarlyStoppingWithLogging(
                monitor=early_stopping_config.monitor,
                patience=early_stopping_config.patience,
                mode="min",
                min_delta=early_stopping_config.min_delta,
            )
        )

    if not any(isinstance(c, ModelCheckpoint) for c in callbacks):
        checkpoint_dir = None if not output_dir else output_dir / "checkpoints"
        callbacks.extend(
            [
                ModelCheckpoint(
                    dirpath=checkpoint_dir,
                    monitor=VALIDATION_LOSS_KEY,
                    filename="best",
                    mode="min",
                    save_top_k=1,
                    save_weights_only=True,
                ),
                ModelCheckpoint(
                    dirpath=checkpoint_dir,
                    filename="last",
                    # ensure that we can load the complete trainer state,
                    # callbacks and all, for resuming later
                    save_weights_only=False,
                ),
            ]
        )

    # special handling for GraphPESCallback: we need to register the
    # output directory with it so that it knows where to save the model etc.
    for cb in callbacks:
        if isinstance(cb, GraphPESCallback):
            cb._register_root(output_dir)
    logger.debug(f"Callbacks: {callbacks}")

    return pl.Trainer(
        **trainer_kwargs,
        logger=lightning_logger,
        callbacks=callbacks,
    )


def main():
    # set the load-atoms verbosity to 1 by default to avoid
    # spamming logs with `rich` output
    os.environ["LOAD_ATOMS_VERBOSE"] = os.getenv("LOAD_ATOMS_VERBOSE", "1")

    # build up the config dict from available sources:
    config_dict = extract_config_dict_from_command_line(
        "Train a GraphPES model using PyTorch Lightning."
    )
    train_from_config(config_dict)


if __name__ == "__main__":
    main()



from __future__ import annotations

from pathlib import Path

import ase.build
import numpy as np
import pytest
import torch
from mattersim.datasets.utils.convertor import GraphConvertor
from mattersim.forcefield.potential import (
    MatterSimCalculator,
    Potential,
    batch_to_dict,
)
from torch_geometric.loader import DataLoader as DataLoader_pyg

from graph_pes.atomic_graph import AtomicGraph, to_batch
from graph_pes.graph_pes_model import GraphPESModel
from graph_pes.interfaces._mattersim import (
    MatterSim_M3Gnet_Wrapper,
    count_number_of_triplets_per_leading_edge,
    mattersim,
)
from graph_pes.utils.calculator import GraphPESCalculator
from graph_pes.utils.lammps import deploy_model
from graph_pes.utils.threebody import (
    triplet_edge_pairs,
)

# Test molecules/crystals
CH4 = ase.build.molecule("CH4")
CH4.center(vacuum=10)
DIAMOND = ase.build.bulk("C", "diamond")
DIAMOND.rattle(0.1)

# models
MATTERSIM_MODEL = Potential.from_checkpoint(  # type: ignore
    "mattersim-v1.0.0-1m", load_training_state=False, device="cpu"
).model
GRAPH_PES_MODEL = MatterSim_M3Gnet_Wrapper(MATTERSIM_MODEL)


def test_output_shapes():
    graph = AtomicGraph.from_ase(DIAMOND)
    outputs = GRAPH_PES_MODEL.get_all_PES_predictions(graph)

    assert outputs["energy"].shape == ()
    assert outputs["forces"].shape == (2, 3)  # 2 atoms in unit cell
    assert outputs["stress"].shape == (3, 3)

    batch = to_batch([graph, graph])
    outputs = GRAPH_PES_MODEL.get_all_PES_predictions(batch)

    assert outputs["energy"].shape == (2,)
    assert outputs["forces"].shape == (4, 3)
    assert outputs["stress"].shape == (2, 3, 3)


@pytest.mark.parametrize("structure", [CH4, DIAMOND])
def test_raw_model_agreement(structure: ase.Atoms):
    graph = AtomicGraph.from_ase(structure)
    us = GRAPH_PES_MODEL.predict_energy(graph)

    c = GraphConvertor(
        twobody_cutoff=GRAPH_PES_MODEL.cutoff.item(),
        threebody_cutoff=GRAPH_PES_MODEL.model.model_args["threebody_cutoff"],
    )
    data = c.convert(structure)
    dl = DataLoader_pyg([data], batch_size=1)
    batch = next(iter(dl))
    batch_dict = batch_to_dict(batch)
    them = MATTERSIM_MODEL(batch_dict)

    assert us.item() == pytest.approx(them.item(), abs=1e-4)


def test_calculator_agreement():
    our_calc = GraphPESCalculator(GRAPH_PES_MODEL)
    our_calc.calculate(DIAMOND, properties=["energy", "forces", "stress"])
    our_results = our_calc.results

    # Compare with the raw model
    ms_calc = MatterSimCalculator(
        Potential.from_checkpoint(  # type: ignore
            "mattersim-v1.0.0-1m", load_training_state=False, device="cpu"
        )
    )
    ms_calc.calculate(DIAMOND, properties=["energy", "forces", "stress"])
    their_results = ms_calc.results
    np.testing.assert_allclose(
        our_results["energy"], their_results["energy"], atol=1e-4
    )
    np.testing.assert_allclose(
        our_results["forces"], their_results["forces"], atol=1e-4
    )
    np.testing.assert_allclose(
        our_results["stress"], their_results["stress"], atol=1e-4
    )


def test_batch_agreement():
    graph = AtomicGraph.from_ase(DIAMOND)
    our_batch = to_batch([graph, graph])
    us = GRAPH_PES_MODEL.predict_energy(our_batch)
    assert us.shape == (2,)

    c = GraphConvertor(
        twobody_cutoff=GRAPH_PES_MODEL.cutoff.item(),
        threebody_cutoff=GRAPH_PES_MODEL.model.model_args["threebody_cutoff"],
    )
    data = c.convert(DIAMOND)
    dl = DataLoader_pyg([data, data], batch_size=2)
    batch = next(iter(dl))
    assert batch.num_graphs == 2
    batch_dict = batch_to_dict(batch)
    them = MATTERSIM_MODEL(batch_dict)
    assert them.shape == (2,)
    torch.testing.assert_close(us, them)

    # test forces
    pot = Potential(MATTERSIM_MODEL)
    their_forces = pot(batch_dict)["forces"]
    assert their_forces.shape == (4, 3)

    our_forces = GRAPH_PES_MODEL.predict_forces(our_batch)
    torch.testing.assert_close(our_forces, their_forces)


def test_api():
    ms = mattersim()
    assert isinstance(ms, GraphPESModel)


def test_implementation():
    c = GraphConvertor(
        twobody_cutoff=GRAPH_PES_MODEL.cutoff.item(),
        threebody_cutoff=GRAPH_PES_MODEL.model.model_args["threebody_cutoff"],
    )
    data = c.convert(DIAMOND)
    dl = DataLoader_pyg([data], batch_size=1)
    batch = next(iter(dl))
    batch_dict = batch_to_dict(batch)

    graph = AtomicGraph.from_ase(DIAMOND)
    graph = graph._replace(
        neighbour_list=batch_dict["edge_index"],
        neighbour_cell_offsets=batch_dict["pbc_offsets"],
    )

    tep = triplet_edge_pairs(graph, GRAPH_PES_MODEL.threebody_cutoff.item())
    assert torch.all(batch_dict["three_body_indices"] == tep).item()

    count = count_number_of_triplets_per_leading_edge(tep, graph).unsqueeze(-1)
    assert torch.all(batch_dict["num_triple_ij"] == count).item()


def test_deploy(tmp_path: Path):
    pytest.skip("TODO: fix threebody torchscripting")

    deploy_model(GRAPH_PES_MODEL, tmp_path / "test_model.pt")



from __future__ import annotations

from typing import Sequence

import torch

from graph_pes.atomic_graph import (
    AtomicGraph,
    PropertyKey,
    has_cell,
    is_batch,
    number_of_atoms,
    number_of_structures,
)
from graph_pes.data.datasets import GraphDataset
from graph_pes.graph_pes_model import GraphPESModel
from graph_pes.utils.misc import all_equal, uniform_repr
from graph_pes.utils.nn import UniformModuleDict


class AdditionModel(GraphPESModel):
    """
    A utility class for combining the predictions of multiple models.

    This is particularly useful for e.g. combining an many-body model with an
    :class:`~graph_pes.models.offsets.EnergyOffset` model to account for the
    arbitrary per-atom energy offsets produced by labelling codes.

    Parameters
    ----------
    models
        the models (given with arbitrary names) to sum.

    Examples
    --------
    Create a model with an explicit offset, two-body and multi-body terms:

    .. code-block:: python

        from graph_pes.models import LennardJones, SchNet, FixedOffset
        from graph_pes.core import AdditionModel

        model = AdditionModel(
            offset=FixedOffset(C=-45.6, H=-1.23),
            pair_wise=LennardJones(cutoff=5.0),
            many_body=SchNet(cutoff=3.0),
        )
    """

    def __init__(self, **models: GraphPESModel):
        max_cutoff = max([m.cutoff.item() for m in models.values()])
        implemented_properties = list(
            set().union(*[m.implemented_properties for m in models.values()])
        )
        super().__init__(
            cutoff=max_cutoff,
            implemented_properties=implemented_properties,
        )
        self.models = UniformModuleDict(**models)

        self.register_buffer(
            "_all_models_same_properties",
            torch.tensor(
                all_equal(
                    [set(m.implemented_properties) for m in models.values()]
                )
            ),
        )

    def predict(
        self, graph: AtomicGraph, properties: list[PropertyKey]
    ) -> dict[PropertyKey, torch.Tensor]:
        device = graph.Z.device
        N = number_of_atoms(graph)

        if is_batch(graph):
            S = number_of_structures(graph)
            zeros = {
                "energy": torch.zeros((S), device=device),
                "forces": torch.zeros((N, 3), device=device),
                "local_energies": torch.zeros((N), device=device),
            }
        else:
            zeros = {
                "energy": torch.zeros((), device=device),
                "forces": torch.zeros((N, 3), device=device),
                "local_energies": torch.zeros((N), device=device),
            }

        if has_cell(graph):
            zeros["stress"] = torch.zeros_like(graph.cell)
            zeros["virial"] = torch.zeros_like(graph.cell)

        final_predictions = {}
        for key in properties:
            final_predictions[key] = zeros[key]

        for model in self.models.values():
            preds = model.predict(graph, properties=properties)
            for key, value in preds.items():
                final_predictions[key] += value

        return final_predictions

    def forward(self, graph: AtomicGraph) -> dict[PropertyKey, torch.Tensor]:
        if not self._all_models_same_properties.item():
            raise ValueError(
                "The forward pass of an AdditionModel is not supported for "
                "models with different implemented properties. "
                "Consider using the predict method instead."
            )

        predictions = [model(graph) for model in self.models.values()]
        return {
            k: torch.stack([p[k] for p in predictions]).sum(dim=0)
            for k in predictions[0]
        }

    def pre_fit_all_components(
        self, graphs: GraphDataset | Sequence[AtomicGraph]
    ):
        for model in self.models.values():
            model.pre_fit_all_components(graphs)

    def __repr__(self):
        return uniform_repr(
            self.__class__.__name__,
            **self.models,
            stringify=True,
            max_width=80,
            indent_width=2,
        )

    def __getitem__(self, key: str) -> GraphPESModel:
        """
        Get a component by name.

        Examples
        --------
        >>> model = AdditionModel(model1=model1, model2=model2)
        >>> model["model1"]
        """
        return self.models[key]
